"""
Utilities
=========

Package-wide utilities.

"""

from collections.abc import Callable, Sequence
from numbers import Number
from operator import itemgetter
from typing import Any

import numpy as np
from tqdm import tqdm as _tqdm


def _check_none(fun: Callable[..., Any]) -> Callable[..., Any | None]:
    """Return None if any input argument is None; Wrapper function."""

    def checked_fun(*args, **kwargs):
        if any(x is None for x in [*args, *(kwargs.values())]):
            return None
        return fun(*args, **kwargs)

    return checked_fun


@_check_none
def res_to_chi2(res: np.ndarray) -> float:
    """Translate residuals to chi2 values, `chi2 = sum(res**2) + C`."""
    return float(np.dot(res, res))


@_check_none
def chi2_to_fval(chi2: float) -> float:
    """Translate chi2 to function value, `fval = 0.5*chi2 = 0.5*sum(res**2) + C`.

    Note that for the function value we thus employ a probabilistic
    interpretation, as the log-likelihood of a standard normal noise model.
    This is in line with e.g. AMICI's and SciPy's objective definition.
    """
    return 0.5 * chi2


@_check_none
def fval_to_chi2(fval: float) -> float:
    """Translate function value to chi2, `chi2 = 2 * fval`.

    Note that for the function value we thus employ a probabilistic
    interpretation, as the log-likelihood of a standard normal noise model.
    This is in line with e.g. AMICI's and SciPy's objective definition.
    """
    return 2.0 * fval


@_check_none
def res_to_fval(res: np.ndarray) -> float:
    """Translate residuals to function value, `fval = 0.5*sum(res**2) + C`."""
    return chi2_to_fval(res_to_chi2(res))


@_check_none
def sres_to_schi2(res: np.ndarray, sres: np.ndarray) -> np.ndarray:
    """Translate residual sensitivities to chi2 gradient."""
    return 2 * res.dot(sres)


@_check_none
def schi2_to_grad(schi2: np.ndarray) -> np.ndarray:
    """Translate chi2 gradient to function value gradient.

    See also :func:`chi2_to_fval`.
    """
    return 0.5 * schi2


@_check_none
def grad_to_schi2(grad: np.ndarray) -> np.ndarray:
    """Translate function value gradient to chi2 gradient."""
    return 2.0 * grad


@_check_none
def sres_to_grad(res: np.ndarray, sres: np.ndarray) -> np.ndarray:
    """Translate residual sensitivities to function value gradient.

    Assumes `fval = 0.5*sum(res**2)`.

    See also :func:`chi2_to_fval`.
    """
    return schi2_to_grad(sres_to_schi2(res, sres))


@_check_none
def sres_to_fim(sres: np.ndarray) -> np.ndarray:
    """Translate residual sensitivities to FIM.

    The FIM is based on the function values, not chi2, i.e. has a normalization
    of 0.5 as in :func:`res_to_fval`.
    """
    return sres.transpose().dot(sres)


def is_none_or_nan(x: Number | None) -> bool:
    """
    Check if x is None or NaN.

    Parameters
    ----------
    x:
        object to be checked

    Returns
    -------
    True if x is None or NaN, False otherwise.
    """
    return x is None or np.isnan(x)


def is_none_or_nan_array(x: Number | np.ndarray | None) -> bool:
    """
    Check if x is None or NaN array.

    Parameters
    ----------
    x:
        object to be checked

    Returns
    -------
    True if x is None or NaN array, False otherwise.
    """
    return x is None or np.isnan(x).all()


def allclose(x: Number | np.ndarray, y: Number | np.ndarray) -> bool:
    """
    Check if two arrays are close.

    Parameters
    ----------
    x: first array
    y: second array

    Returns
    -------
    True if all elements of x and y are close, False otherwise.
    """
    # Note: We use this wrapper around np.allclose in order to more easily
    #  adjust hyper parameters for the tolerance.
    return np.allclose(x, y)


def isclose(
    x: Number | np.ndarray,
    y: Number | np.ndarray,
) -> bool | np.ndarray:
    """
    Check if two values or arrays are close, element-wise.

    Parameters
    ----------
    x: first array
    y: second array

    Returns
    -------
    Element-wise boolean comparison of x and y.
    """
    # Note: We use this wrapper around np.isclose in order to more easily
    #  adjust hyper parameters for the tolerance.
    return np.isclose(x, y)


def get_condition_label(condition_id: str) -> str:
    """Convert a condition ID to a label.

    Labels for conditions are used at different locations (e.g. ensemble
    prediction code, and visualization code). This method ensures that the same
    condition is labeled identically everywhere.

    Parameters
    ----------
    condition_id:
        The condition ID that will be used to generate a label.

    Returns
    -------
    The condition label.
    """
    return f"condition_{condition_id}"


def assign_clusters(vals):
    """
    Find clustering.

    Parameters
    ----------
    vals: numeric list or array
        List to be clustered.

    Returns
    -------
    clust: numeric list
        Indicating the corresponding cluster of each element from
        'vals'.
    clustsize: numeric list
        Size of clusters, length equals number of clusters.
    """
    # sanity checks
    if vals is None or len(vals) == 0:
        return [], []
    elif len(vals) == 1:
        return np.array([0]), np.array([1.0])

    # sort values
    vals = np.sort(vals)

    # assign cluster to first element and then assign new cluster every time distance is greater 0.1
    cluster_indices = np.append([0], np.cumsum(np.diff(vals) > 0.1))

    # get cluster sizes
    _, cluster_size = np.unique(cluster_indices, return_counts=True)
    return cluster_indices, cluster_size


def delete_nan_inf(
    fvals: np.ndarray,
    x: Sequence[np.ndarray | list[float]] | None = None,
    xdim: int | None = 1,
    magnitude_bound: float | None = np.inf,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Delete nan and inf values in fvals.

    If parameters 'x' are passed, also the corresponding entries are deleted.

    Parameters
    ----------
    x:
        array of parameters
    fvals:
        array of fval
    xdim:
        dimension of x, in case x dimension cannot be inferred
    magnitude_bound:
        any values with a magnitude (absolute value) larger than the
        `magnitude_bound` are also deleted

    Returns
    -------
    x:
        array of parameters without nan or inf
    fvals:
        array of fval without nan or inf
    """
    fvals = np.asarray(fvals)
    finite_fvals = np.isfinite(fvals) & (np.absolute(fvals) < magnitude_bound)
    if x is not None:
        if np.isfinite(fvals).any():
            finite_indices = np.where(finite_fvals)[0]
            x = np.array(itemgetter(*finite_indices)(x))
            # itemgetter does not return list for single element
            if len(x.shape) == 1:
                x = x.reshape((1, x.shape[0]))
        else:
            # If none of the fvals are finite we try to construct an
            # empty np.ndarray with the correct dimension (other functions rely
            # on x.shape[1] to be of correct dimension)
            x = np.array(x)
            x = np.empty(
                (
                    0,
                    (
                        x.shape[1]
                        if x.ndim == 2
                        else x[0].shape[0]
                        if x[0] is not None
                        else xdim
                    ),
                )
            )
    return x, fvals[finite_fvals]


def tqdm(*args, enable: bool = None, **kwargs):
    """
    Create a progress bar using tqdm.

    Parameters
    ----------
    args:
        Arguments passed to tqdm.
    enable:
        Whether to enable the progress bar.
        If None, use tqdm defaults.
        Mutually exclusive with `disable`.
    kwargs:
        Keyword arguments passed to tqdm.

    Returns
    -------
    progress_bar:
        A progress bar.
    """
    # Drop the `disable` argument unless it is not-None.
    # This way, we don't interfere with TQDM_DISABLE or other global
    # tqdm settings.
    disable = kwargs.pop("disable", None)

    if enable is not None:
        if disable is not None and enable != disable:
            raise ValueError(
                "Contradicting values for `enable` and `disable` passed."
            )
        disable = not enable

    if disable is not None:
        kwargs["disable"] = disable
    return _tqdm(*args, **kwargs)



"""Module for the InnerCalculatorCollector class.

In case of semi-quantitative or qualitative measurements, this class is used
to collect hierarchical inner calculators for each data type and merge their results.
"""

from __future__ import annotations

import copy
import warnings
from collections.abc import Sequence
from typing import Union

import numpy as np

from ..C import (
    AMICI_SIGMAY,
    AMICI_SSIGMAY,
    AMICI_SSIGMAZ,
    AMICI_SY,
    AMICI_Y,
    CENSORED,
    FVAL,
    GRAD,
    HESS,
    INNER_PARAMETERS,
    METHOD,
    MODE_RES,
    ORDINAL,
    ORDINAL_OPTIONS,
    RDATAS,
    RELATIVE,
    RES,
    SEMIQUANTITATIVE,
    SPLINE_APPROXIMATION_OPTIONS,
    SPLINE_KNOTS,
    SPLINE_RATIO,
    SRES,
    ModeType,
)
from ..objective.amici.amici_calculator import AmiciCalculator
from ..objective.amici.amici_util import (
    add_sim_grad_to_opt_grad,
    filter_return_dict,
    init_return_values,
)

try:
    import amici.sim.sundials as asd
    import petab.v1 as petab
    from amici.sim._parameter_mapping import ParameterMapping
except ImportError:
    petab = None
    ParameterMapping = None

from .ordinal import OrdinalCalculator, OrdinalInnerSolver, OrdinalProblem
from .relative import RelativeAmiciCalculator, RelativeInnerProblem
from .semiquantitative import (
    SemiquantCalculator,
    SemiquantInnerSolver,
    SemiquantProblem,
)

AmiciModel = Union["asd.Model", "asd.ModelPtr"]
AmiciSolver = Union["asd.Solver", "asd.SolverPtr"]


class InnerCalculatorCollector(AmiciCalculator):
    """Class to collect inner calculators in case of non-quantitative data types.

    Upon import of a petab problem, the PEtab importer checks whether there are
    non-quantitative data types. If so, it creates an instance of this class
    instead of an AmiciCalculator. This class then collects the inner calculators
    for each data type and merges their results with the quantitative results.

    Parameters
    ----------
    data_types:
        List of non-quantitative data types in the problem.
    petab_problem:
        The PEtab problem.
    model:
        The AMICI model.
    edatas:
        The experimental data.
    inner_options:
        Options for the inner problems and solvers.
    """

    def __init__(
        self,
        data_types: set[str],
        petab_problem: petab.Problem,
        model: AmiciModel,
        edatas: list[asd.ExpData],
        inner_options: dict,
    ):
        super().__init__()
        self.validate_options(inner_options)

        self.data_types = data_types
        self.inner_calculators: list[
            AmiciCalculator
        ] = []  # TODO make into a dictionary (future PR, together with .hierarchical of Problem)

        self.semiquant_observable_ids = None
        self.relative_observable_ids = None

        self.construct_inner_calculators(
            petab_problem, model, edatas, inner_options
        )

        self.quantitative_data_mask = self._get_quantitative_data_mask(edatas)

        self._known_least_squares_safe = False

    def initialize(self):
        """Initialize."""
        for calculator in self.inner_calculators:
            calculator.initialize()

    def construct_inner_calculators(
        self,
        petab_problem: petab.Problem,
        model: AmiciModel,
        edatas: list[asd.ExpData],
        inner_options: dict,
    ):
        """Construct inner calculators for each data type."""
        self.necessary_par_dummy_values = {}

        if RELATIVE in self.data_types:
            relative_inner_problem = RelativeInnerProblem.from_petab_amici(
                petab_problem, model, edatas
            )
            self.necessary_par_dummy_values.update(
                relative_inner_problem.get_dummy_values(scaled=True)
            )
            relative_inner_solver = RelativeAmiciCalculator(
                inner_problem=relative_inner_problem
            )
            self.inner_calculators.append(relative_inner_solver)
            self.relative_observable_ids = (
                relative_inner_problem.get_relative_observable_ids()
            )

        if ORDINAL in self.data_types or CENSORED in self.data_types:
            optimal_scaling_inner_options = {
                key: value
                for key, value in inner_options.items()
                if key in ORDINAL_OPTIONS
            }
            inner_problem_method = optimal_scaling_inner_options.get(
                METHOD, None
            )
            ordinal_inner_problem = OrdinalProblem.from_petab_amici(
                petab_problem, model, edatas, inner_problem_method
            )
            ordinal_inner_solver = OrdinalInnerSolver(
                options=optimal_scaling_inner_options
            )
            ordinal_calculator = OrdinalCalculator(
                ordinal_inner_problem, ordinal_inner_solver
            )
            self.inner_calculators.append(ordinal_calculator)

        if SEMIQUANTITATIVE in self.data_types:
            spline_inner_options = {
                key: value
                for key, value in inner_options.items()
                if key in SPLINE_APPROXIMATION_OPTIONS
            }
            spline_ratio = spline_inner_options.pop(SPLINE_RATIO, None)
            semiquant_problem = SemiquantProblem.from_petab_amici(
                petab_problem, model, edatas, spline_ratio
            )
            semiquant_inner_solver = SemiquantInnerSolver(
                options=spline_inner_options
            )
            semiquant_calculator = SemiquantCalculator(
                semiquant_problem, semiquant_inner_solver
            )
            self.necessary_par_dummy_values.update(
                semiquant_problem.get_noise_dummy_values(scaled=True)
            )
            self.inner_calculators.append(semiquant_calculator)
            self.semiquant_observable_ids = (
                semiquant_problem.get_semiquant_observable_ids()
            )

        if self.data_types - {
            RELATIVE,
            ORDINAL,
            CENSORED,
            SEMIQUANTITATIVE,
        }:
            unsupported_data_types = self.data_types - {
                RELATIVE,
                ORDINAL,
                CENSORED,
                SEMIQUANTITATIVE,
            }
            raise NotImplementedError(
                f"Data types {unsupported_data_types} are not supported."
            )

    def validate_options(self, inner_options: dict):
        """Validate the inner options.

        Parameters
        ----------
        inner_options:
            Options for the inner problems and solvers.
        """
        for key in inner_options:
            if (
                key not in ORDINAL_OPTIONS
                and key not in SPLINE_APPROXIMATION_OPTIONS
            ):
                raise ValueError(f"Unknown inner option {key}.")

    def _get_quantitative_data_mask(
        self,
        edatas: list[asd.ExpData],
    ) -> list[np.ndarray]:
        # transform experimental data
        edatas = [asd.ExpDataView(edata)["measurements"] for edata in edatas]

        quantitative_data_mask = [
            np.ones_like(edata, dtype=bool) for edata in edatas
        ]

        # iterate over inner problems
        for calculator in self.inner_calculators:
            inner_parameters = calculator.inner_problem.xs.values()
            # Remove inner parameter masks from quantitative data mask
            for inner_par in inner_parameters:
                for cond_idx, condition_mask in enumerate(
                    quantitative_data_mask
                ):
                    condition_mask[inner_par.ixs[cond_idx]] = False

        # Put to False all entries that have a nan value in the edata
        for condition_mask, edata in zip(
            quantitative_data_mask, edatas, strict=True
        ):
            condition_mask[np.isnan(edata)] = False

        # If there is no quantitative data, return None
        if not all(mask.any() for mask in quantitative_data_mask):
            return None

        return quantitative_data_mask

    def get_inner_par_ids(self) -> list[str]:
        """Return the ids of inner parameters of all inner problems."""
        return [
            parameter_id
            for inner_calculator in self.inner_calculators
            for parameter_id in inner_calculator.inner_problem.get_x_ids()
        ]

    def get_interpretable_inner_par_ids(self) -> list[str]:
        """Return the ids of interpretable inner parameters of all inner problems.

        See :func:`InnerProblem.get_interpretable_x_ids`.
        """
        return [
            parameter_id
            for inner_calculator in self.inner_calculators
            for parameter_id in inner_calculator.inner_problem.get_interpretable_x_ids()
        ]

    def get_interpretable_inner_par_bounds(
        self,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Return the bounds of interpretable inner parameters of all inner problems."""
        lb = []
        ub = []
        for inner_calculator in self.inner_calculators:
            (
                lb_i,
                ub_i,
            ) = inner_calculator.inner_problem.get_interpretable_x_bounds()
            lb.extend(lb_i)
            ub.extend(ub_i)
        return np.asarray(lb), np.asarray(ub)

    def get_interpretable_inner_par_scales(self) -> list[str]:
        """Return the scales of interpretable inner parameters of all inner problems."""
        return [
            scale
            for inner_calculator in self.inner_calculators
            for scale in inner_calculator.inner_problem.get_interpretable_x_scales()
        ]

    def __call__(
        self,
        x_dct: dict,
        sensi_orders: tuple[int],
        mode: ModeType,
        amici_model: AmiciModel,
        amici_solver: AmiciSolver,
        edatas: list[asd.ExpData],
        n_threads: int,
        x_ids: Sequence[str],
        parameter_mapping: ParameterMapping,
        fim_for_hess: bool,
    ):
        """Perform the actual AMICI call.

        Called within the :func:`AmiciObjective.__call__` method.
        Calls all the inner calculators and combines the results.

        Parameters
        ----------
        x_dct:
            Parameters for which to compute function value and derivatives.
        sensi_orders:
            Tuple of requested sensitivity orders.
        mode:
            Call mode (function value or residual based).
        amici_model:
            The AMICI model.
        amici_solver:
            The AMICI solver.
        edatas:
            The experimental data.
        n_threads:
            Number of threads for AMICI call.
        x_ids:
            Ids of optimization parameters.
        parameter_mapping:
            Mapping of optimization to simulation parameters.
        fim_for_hess:
            Whether to use the FIM (if available) instead of the Hessian (if
            requested).
        """
        from amici.sim.sundials.petab.v1 import fill_in_parameters

        if mode == MODE_RES and any(
            data_type in self.data_types
            for data_type in [ORDINAL, CENSORED, SEMIQUANTITATIVE]
        ):
            raise NotImplementedError(
                f"Mode {mode} is not implemented for ordinal, censored or semi-quantitative data. "
                "However, it can be used if the only non-quantitative data type is relative data."
            )

        if 2 in sensi_orders and any(
            data_type in self.data_types
            for data_type in [ORDINAL, CENSORED, SEMIQUANTITATIVE]
        ):
            raise ValueError(
                "Hessian and FIM are not implemented for ordinal, censored or semi-quantitative data. "
                "However, they can be used if the only non-quantitative data type is relative data."
            )

        if (
            amici_solver.get_sensitivity_method()
            == asd.SensitivityMethod.adjoint
            and any(
                data_type in self.data_types
                for data_type in [ORDINAL, CENSORED, SEMIQUANTITATIVE]
            )
        ):
            raise NotImplementedError(
                "Adjoint sensitivity analysis is not implemented for ordinal, censored or semi-quantitative data. "
                "However, it can be used if the only non-quantitative data type is relative data."
            )

        # if we're using adjoint sensitivity analysis or need second order
        # sensitivities or are in residual mode, we can do so if the only
        # non-quantitative data type is relative data. In this case, we
        # use the relative calculator directly.
        if (
            amici_solver.get_sensitivity_method()
            == asd.SensitivityMethod.adjoint
            or 2 in sensi_orders
            or mode == MODE_RES
        ):
            relative_calculator = self.inner_calculators[0]
            ret = relative_calculator(
                x_dct=x_dct,
                sensi_orders=sensi_orders,
                mode=mode,
                amici_model=amici_model,
                amici_solver=amici_solver,
                edatas=edatas,
                n_threads=n_threads,
                x_ids=x_ids,
                parameter_mapping=parameter_mapping,
                fim_for_hess=fim_for_hess,
            )
            return filter_return_dict(ret)

        # get dimension of outer problem
        dim = len(x_ids)

        # initialize return values
        nllh, snllh, s2nllh, chi2, res, sres = init_return_values(
            sensi_orders, mode, dim
        )
        spline_knots = None
        interpretable_inner_pars = []

        # set order in solver
        sensi_order = 0
        if sensi_orders:
            sensi_order = max(sensi_orders)

        amici_solver.set_sensitivity_order(sensi_order)

        x_dct = copy.deepcopy(x_dct)
        x_dct.update(self.necessary_par_dummy_values)
        # fill in parameters, we expect here a RunTimeWarning to occur
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message="The following problem parameters were not used:.*",
                category=RuntimeWarning,
            )
            fill_in_parameters(
                edatas=edatas,
                problem_parameters=x_dct,
                scaled_parameters=True,
                parameter_mapping=parameter_mapping,
                amici_model=amici_model,
            )

        # run amici simulation
        rdatas = asd.run_simulations(
            amici_model,
            amici_solver,
            edatas,
            num_threads=min(n_threads, len(edatas)),
        )

        # if any amici simulation failed, it's unlikely we can compute
        # meaningful inner parameters, so we better just fail early.
        if any(rdata.status != asd.AMICI_SUCCESS for rdata in rdatas):
            ret = {
                FVAL: nllh,
                GRAD: snllh,
                HESS: s2nllh,
                RES: res,
                SRES: sres,
                RDATAS: rdatas,
                SPLINE_KNOTS: None,
                INNER_PARAMETERS: None,
            }
            ret[FVAL] = np.inf
            # if the gradient was requested,
            # we need to provide some value for it
            if 1 in sensi_orders:
                ret[GRAD] = np.full(shape=len(x_ids), fill_value=np.nan)
            return filter_return_dict(ret)

        if (
            not self._known_least_squares_safe
            and mode == MODE_RES
            and 1 in sensi_orders
        ):
            if not amici_model.get_add_sigma_residuals() and any(
                (
                    (r[AMICI_SSIGMAY] is not None and np.any(r[AMICI_SSIGMAY]))
                    or (
                        r[AMICI_SSIGMAZ] is not None
                        and np.any(r[AMICI_SSIGMAZ])
                    )
                )
                for r in rdatas
            ):
                raise RuntimeError(
                    "Cannot use least squares solver with"
                    "parameter dependent sigma! Support can be "
                    "enabled via "
                    "amici_model.setAddSigmaResiduals()."
                )
            self._known_least_squares_safe = True  # don't check this again

        # call inner calculators and collect results
        for calculator in self.inner_calculators:
            inner_result = calculator(
                x_dct=x_dct,
                sensi_orders=sensi_orders,
                mode=mode,
                amici_model=amici_model,
                amici_solver=amici_solver,
                edatas=edatas,
                n_threads=n_threads,
                x_ids=x_ids,
                parameter_mapping=parameter_mapping,
                fim_for_hess=fim_for_hess,
                rdatas=rdatas,
            )
            nllh += inner_result[FVAL]
            if 1 in sensi_orders:
                snllh += inner_result[GRAD]

            inner_pars = inner_result.get(INNER_PARAMETERS)
            if inner_pars is not None:
                interpretable_inner_pars.extend(inner_pars)
            if SPLINE_KNOTS in inner_result:
                spline_knots = inner_result[SPLINE_KNOTS]

        # add the quantitative data contribution
        if self.quantitative_data_mask is not None:
            quantitative_result = calculate_quantitative_result(
                rdatas=rdatas,
                sensi_orders=sensi_orders,
                edatas=edatas,
                mode=mode,
                quantitative_data_mask=self.quantitative_data_mask,
                dim=dim,
                parameter_mapping=parameter_mapping,
                par_opt_ids=x_ids,
                par_sim_ids=amici_model.get_free_parameter_ids(),
            )
            nllh += quantitative_result[FVAL]
            if 1 in sensi_orders:
                snllh += quantitative_result[GRAD]

        ret = {
            FVAL: nllh,
            GRAD: snllh,
            HESS: s2nllh,
            RES: res,
            SRES: sres,
            RDATAS: rdatas,
        }

        ret[INNER_PARAMETERS] = (
            interpretable_inner_pars
            if len(interpretable_inner_pars) > 0
            else None
        )
        ret[SPLINE_KNOTS] = spline_knots

        return filter_return_dict(ret)


def calculate_quantitative_result(
    rdatas: list[asd.ReturnDataView],
    edatas: list[asd.ExpData],
    sensi_orders: tuple[int],
    mode: ModeType,
    quantitative_data_mask: list[np.ndarray],
    dim: int,
    parameter_mapping: ParameterMapping,
    par_opt_ids: list[str],
    par_sim_ids: list[str],
):
    """Calculate the function values from rdatas and return as dict."""
    nllh, snllh, s2nllh, chi2, res, sres = init_return_values(
        sensi_orders, mode, dim
    )

    # transform experimental data
    edatas = [asd.ExpDataView(edata)["measurements"] for edata in edatas]

    # calculate the function value
    for rdata, edata, mask in zip(
        rdatas, edatas, quantitative_data_mask, strict=True
    ):
        data_i = edata[mask]
        sim_i = rdata[AMICI_Y][mask]
        sigma_i = rdata[AMICI_SIGMAY][mask]

        nllh += 0.5 * np.nansum(
            np.log(2 * np.pi * sigma_i**2) + (data_i - sim_i) ** 2 / sigma_i**2
        )

    # calculate the gradient if requested
    if 1 in sensi_orders:
        parameter_map_sim_var = [
            cond_par_map.map_sim_var for cond_par_map in parameter_mapping
        ]
        # iterate over simulation conditions
        for (
            rdata,
            edata,
            mask,
            condition_map_sim_var,
        ) in zip(
            rdatas,
            edatas,
            quantitative_data_mask,
            parameter_map_sim_var,
            strict=True,
        ):
            data_i = edata[mask]
            sim_i = rdata[AMICI_Y][mask]
            sigma_i = rdata[AMICI_SIGMAY][mask]

            n_parameters = rdata[AMICI_SY].shape[1]

            # Get sensitivities of observables and sigmas
            sensitivities_i = np.asarray(
                [
                    rdata[AMICI_SY][:, parameter_index, :][mask]
                    for parameter_index in range(n_parameters)
                ]
            )
            ssigma_i = np.asarray(
                [
                    rdata[AMICI_SSIGMAY][:, parameter_index, :][mask]
                    for parameter_index in range(n_parameters)
                ]
            )
            # calculate the gradient for the condition
            gradient_for_condition = np.nansum(
                np.multiply(
                    ssigma_i,
                    (
                        (
                            np.full(len(data_i), 1)
                            - (data_i - sim_i) ** 2 / sigma_i**2
                        )
                        / sigma_i
                    ),
                ),
                axis=1,
            ) + np.nansum(
                np.multiply(sensitivities_i, ((sim_i - data_i) / sigma_i**2)),
                axis=1,
            )
            add_sim_grad_to_opt_grad(
                par_opt_ids=par_opt_ids,
                par_sim_ids=par_sim_ids,
                condition_map_sim_var=condition_map_sim_var,
                sim_grad=gradient_for_condition,
                opt_grad=snllh,
            )

    ret = {
        FVAL: nllh,
        GRAD: snllh,
        HESS: s2nllh,
        RES: res,
        SRES: sres,
        RDATAS: rdatas,
    }
    return filter_return_dict(ret)



"""
Result
======

The pypesto.Result object contains all results generated by
the pypesto components. It contains sub-results for
optimization, profiling, sampling.
"""

from .optimize import LazyOptimizerResult, OptimizeResult, OptimizerResult
from .predict import PredictionConditionResult, PredictionResult
from .profile import ProfileResult, ProfilerResult
from .result import Result
from .sample import McmcPtResult, SampleResult



"""
Model Selection
===============

Perform model selection with a
`PEtab Select <https://petab-select.readthedocs.io/>`_ problem.
"""

from . import postprocessors
from .misc import SacessMinimizeMethod, model_to_pypesto_problem
from .model_problem import ModelProblem
from .problem import Problem

try:
    import petab_select

    del petab_select
except ImportError:
    import warnings

    warnings.warn(
        "pyPESTO's model selection methods require an installation of PEtab "
        "Select (https://github.com/PEtab-dev/petab_select). Install via "
        "`pip3 install petab-select` or `pip3 install pypesto[select]`.",
        stacklevel=1,
    )



"""
Semi-quantitative data integration
==================================

Contains the implementation of a spline approximation approach, applied for
integration of semi-quantitative data in ODE modeling, where the data is
assumed to have an unknown monotone relationship with the model output. This
relationship is approximated by a piecewise linear (spline) function, which
is numerically optimized to fit the data. This constitutes the inner subproblem
of the hierarchical optimization problem.

An example of parameter estimation with semi-quantitative data
can be found in pypesto/doc/examples/semiquantitative_data.ipynb.
"""

from .calculator import SemiquantCalculator
from .parameter import SplineInnerParameter
from .problem import SemiquantProblem
from .solver import SemiquantInnerSolver



"""
Visualization of the model fit after optimization.

Currently only for PEtab problems.
"""

import copy
import logging
from collections.abc import Sequence
from typing import Union

import amici.sim.sundials as asd
import matplotlib.axes
import matplotlib.pyplot as plt
import numpy as np
import petab.v1 as petab
from amici.sim.sundials.petab.v1 import (
    fill_in_parameters,
    rdatas_to_simulation_df,
)
from amici.sim.sundials.plotting import (
    plot_observable_trajectories,
    plot_state_trajectories,
)
from petab.v1.visualize import plot_problem

from ..C import CENSORED, ORDINAL, RDATAS, SEMIQUANTITATIVE
from ..objective import (
    AggregatedObjective,
    AmiciObjective,
    NegLogParameterPriors,
    NegLogPriors,
)
from ..petab.importer import get_petab_non_quantitative_data_types
from ..problem import HierarchicalProblem, Problem
from ..result import Result
from .observable_mapping import _add_spline_mapped_simulations_to_model_fit
from .ordinal_categories import plot_categories_from_pypesto_result

AmiciModel = Union["asd.Model", "asd.ModelPtr"]

__all__ = ["visualize_optimized_model_fit", "time_trajectory_model"]

logger = logging.getLogger(__name__)


def visualize_optimized_model_fit(
    petab_problem: petab.Problem,
    result: Result | Sequence[Result],
    pypesto_problem: Problem,
    start_index: int = 0,
    return_dict: bool = False,
    unflattened_petab_problem: petab.Problem = None,
    **kwargs,
) -> matplotlib.axes.Axes | dict | list[matplotlib.axes.Axes] | list[dict]:
    """
    Visualize the optimized model fit of a PEtab problem.

    Function calls the PEtab visualization file of the ``petab_problem`` and
    visualizes the fit of the optimized parameter. Common additional
    argument is ``subplot_dir`` to specify the directory each subplot is
    saved to. Further keyword arguments are delegated to
    :func:`petab.visualize.plot_with_vis_spec`, see there for more information.

    Parameters
    ----------
    petab_problem:
        The :py:class:`petab.Problem` that was optimized.
    result:
        The result object from optimization.
    start_index:
        The index of the optimization run in `result.optimize_result.list`.
        Ignored if `problem_parameters` is provided.
    pypesto_problem:
        The pyPESTO problem.
    return_dict:
        Return plot and simulation results as a dictionary.
    unflattened_petab_problem:
        If the original PEtab problem is flattened, this can be passed
        to plot with the original unflattened problem.
    kwargs:
        Passed to :func:`petab.visualize.plot_problem`.

    Returns
    -------
    axes: `matplotlib.axes.Axes` object of the created plot.
    None: In case subplots are saved to file
    """

    amici_models, objective_results = [], []

    # get amici model
    if isinstance(pypesto_problem.objective, AmiciObjective):
        x = result.optimize_result.list[start_index]["x"][
            pypesto_problem.x_free_indices
        ]

        objective_results.append(
            pypesto_problem.objective(x, return_dict=True)
        )
        amici_models.append(pypesto_problem.objective.amici_model)
    elif isinstance(pypesto_problem.objective, AggregatedObjective):
        # iterate over objectives to find the amici models in the AggregatedObjective
        x = result.optimize_result.list[start_index]["x"]

        for objective in pypesto_problem.objective._objectives:
            if isinstance(objective, AmiciObjective):
                amici_models.append(objective.amici_model)
                # objective expects full vector here
                objective_results.append(objective(x, return_dict=True))
            elif isinstance(objective, NegLogParameterPriors | NegLogPriors):
                # priors are not used for simulation
                pass
            else:
                logger.warning(
                    f"Objective {objective} is not an 'AmiciObjective' "
                    f"or a 'NegLogParameterPriors'/'NegLogPriors' and hence will be ignored."
                )

        if len(amici_models) == 0:
            raise ValueError(
                "'visualize_optimized_model_fit' only works with amici models. None were found."
            )
    else:
        raise ValueError(
            "'visualize_optimized_model_fit' only works with an 'AmiciObjective' or"
            " an 'AggregatedObjective' containing an 'AmiciObjective'."
        )

    axes_list, simulation_df_list = [], []
    for amici_model, objective_result in zip(
        amici_models, objective_results, strict=True
    ):
        # get simulation results
        simulation_df = rdatas_to_simulation_df(
            objective_result[RDATAS],
            amici_model,
            petab_problem.measurement_df,  # todo: potentially needs to be changed to work with multiple models
        )

        # handle flattened PEtab problems
        petab_problem_to_plot = petab_problem
        if unflattened_petab_problem:
            simulation_df = petab.core.unflatten_simulation_df(
                simulation_df=simulation_df,
                petab_problem=unflattened_petab_problem,
            )
            petab_problem_to_plot = unflattened_petab_problem

        # plot
        axes = plot_problem(
            petab_problem=petab_problem_to_plot,
            simulations_df=simulation_df,
            **kwargs,
        )

        non_quantitative_data_types = get_petab_non_quantitative_data_types(
            petab_problem
        )

        if non_quantitative_data_types is not None:
            if (
                ORDINAL in non_quantitative_data_types
                or CENSORED in non_quantitative_data_types
            ):
                axes = plot_categories_from_pypesto_result(
                    result,
                    start_index=start_index,
                    axes=axes,
                )
            if SEMIQUANTITATIVE in non_quantitative_data_types:
                axes = _add_spline_mapped_simulations_to_model_fit(
                    result=result,
                    pypesto_problem=pypesto_problem,
                    start_index=start_index,
                    axes=axes,
                )

        axes_list.append(axes)
        simulation_df_list.append(simulation_df)

    if return_dict:
        dict_list = []
        for axes, simulation_df, objective_result in zip(
            axes_list, simulation_df_list, objective_results, strict=True
        ):
            dict_list.append(
                {
                    "axes": axes,
                    "objective_result": objective_result,
                    "simulation_df": simulation_df,
                }
            )
        if len(dict_list) == 1:
            return dict_list[0]
        return dict_list
    if len(axes_list) == 1:
        return axes_list[0]
    return axes_list


def time_trajectory_model(
    result: Result | Sequence[Result],
    problem: Problem = None,
    # TODO: conditions: Union[str, Sequence[str]] = None,
    timepoints: np.ndarray | Sequence[np.ndarray] = None,
    n_timepoints: int = 1000,
    start_index: int = 0,
    state_ids: str | Sequence[str] = None,
    state_names: str | Sequence[str] = None,
    observable_ids: str | Sequence[str] = None,
) -> matplotlib.axes.Axes | None:
    """
    Visualize the time trajectory of the model with given timepoints.

    It does this by calling the amici plotting routines.

    Parameters
    ----------
    result:
        The result object from optimization.
    problem:
        A pypesto problem. Default is 'None' in which case result.problem is
        used. Needed in case the result is loaded from hdf5.
    timepoints:
        Array of timepoints, at which the trajectory will be plotted.
    n_timepoints:
        Number of timepoints to be plotted between 0 and last measurement of
        the model. Only used when timepoints==None.
    start_index:
        Index of Optimization run to be plotted. Default is best start.
    state_ids:
        Ids of the states to be plotted.
    state_names:
        Names of the states to be plotted.
    observable_ids:
        Ids of the observables to be plotted.

    Returns
    -------
    axes:
        `matplotlib.axes.Axes` object of the plot.
    """
    if problem is None:
        problem = result.problem
    # add timepoints as needed
    if timepoints is None:
        end_time = max(problem.objective.edatas[0].get_timepoints())
        timepoints = np.linspace(start=0, stop=end_time, num=n_timepoints)

    # get rdatas
    rdatas = _get_simulation_rdatas(
        result=result,
        problem=problem,
        start_index=start_index,
        simulation_timepoints=timepoints,
    )

    if state_ids == [] and state_names == []:
        axes = _time_trajectory_model_without_states(
            model=problem.objective.amici_model,
            rdatas=rdatas,
            observable_ids=observable_ids,
        )
    else:
        axes = _time_trajectory_model_with_states(
            model=problem.objective.amici_model,
            rdatas=rdatas,
            state_ids=state_ids,
            state_names=state_names,
            observable_ids=observable_ids,
        )

    return axes


def _get_simulation_rdatas(
    result: Result | Sequence[Result],
    problem: Problem,
    start_index: int = 0,
    simulation_timepoints: np.ndarray = None,
) -> list[asd.ReturnData]:
    """
    Get simulation results for a given optimization result and timepoints.

    Parameters
    ----------
    result:
        The result object from optimization.
    problem:
        A pypesto problem.
    timepoints:
        Array of timepoints, at which the trajectory will be stored..

    start_index:
        Index of Optimization run to be plotted. Default is best start.
    Returns
    -------
    rdatas:
        List of amici.ReturnData objects containing the simulation results.
    """
    # add timepoints as needed
    if simulation_timepoints is None:
        end_time = max(problem.objective.edatas[0].getTimepoints())
        simulation_timepoints = np.linspace(start=0, stop=end_time, num=1000)

    # get optimization result
    parameters = result.optimize_result.list[start_index]["x"]

    # reduce vector to only include free indices. Needed downstream.
    parameters = problem.get_reduced_vector(parameters)

    # simulate with custom timepoints for hierarchical model
    if isinstance(problem, HierarchicalProblem):
        # get parameter dictionary
        x_dct = dict(
            zip(
                problem.x_names,
                result.optimize_result.list[start_index].x,
                strict=True,
            )
        )

        # evaluate objective with return dict = True to get inner parameters
        ret = problem.objective(
            parameters, mode="mode_fun", sensi_orders=(0,), return_dict=True
        )

        # update parameter dictionary with inner parameters
        inner_parameter_dict = dict(
            zip(problem.inner_x_names, ret["inner_parameters"], strict=True)
        )
        x_dct.update(inner_parameter_dict)

        parameter_mapping = problem.objective.parameter_mapping
        edatas = copy.deepcopy(problem.objective.edatas)
        amici_model = problem.objective.amici_model

        # disable sensitivities to improve computation time
        amici_solver = copy.deepcopy(problem.objective.amici_solver)
        amici_solver.setSensitivityOrder(asd.SensitivityOrder.none)

        for j in range(len(edatas)):
            edatas[j].setTimepoints(simulation_timepoints)

        fill_in_parameters(
            edatas=edatas,
            problem_parameters=x_dct,
            scaled_parameters=True,
            parameter_mapping=parameter_mapping,
            amici_model=amici_model,
        )

        rdatas = asd.run_simulations(
            amici_model,
            amici_solver,
            edatas,
        )
    else:
        # set custom timepoints
        obj = problem.objective.set_custom_timepoints(
            timepoints_global=simulation_timepoints
        )

        # evaluate objective with return dict = True to get data
        ret = obj(
            parameters, mode="mode_fun", sensi_orders=(0,), return_dict=True
        )
        rdatas = ret["rdatas"]

    return rdatas


def _time_trajectory_model_with_states(
    model: AmiciModel,
    rdatas: Union["asd.ReturnData", Sequence["asd.ReturnData"]],
    state_ids: Sequence[str],
    state_names: Sequence[str],
    observable_ids: str | Sequence[str],
):
    """
    Visualizes both, states and observables.

    Helper function for time_trajectory_model.

    Parameters
    ----------
    model:
        The amici.Model from the model of interest. Used to annotate the plot.
    rdatas:
        The data to be plotted. Each entry in the Sequence corresponds to a
        condition.
    state_ids:
        Ids of the states to be plotted.
    state_names:
        Names of the states to be plotted.
    observable_ids:
        Ids of the observable Ids to be plotted.

    Returns
    -------
    axes:
        `matplotlib.axes.Axes` object of the plot.
    """
    # if state_name, state_id or observable_id is not None, get indices
    # for these the AMICI plotting functions default to all indices if
    # `None` is specified.
    state_indices_by_id = []
    state_indices_by_name = []
    if state_ids is not None:
        state_indices_by_id = [
            model.get_state_ids().index(state_id) for state_id in state_ids
        ]
    if state_names is not None:
        state_indices_by_name = [
            model.getStateNames().index(state_name)
            for state_name in state_names
        ]
    state_indices = list(set(state_indices_by_id + state_indices_by_name))
    if state_indices == []:
        state_indices = None

    observable_indices = None
    if observable_ids is not None:
        observable_indices = [
            model.get_observable_ids().index(obs_id)
            for obs_id in observable_ids
        ]

    fig, axes = plt.subplots(len(rdatas), 2)
    # enforce two dimensions in case there is only one condition
    axes = np.atleast_2d(axes)

    for i_cond, rdata in enumerate(rdatas):
        plot_state_trajectories(
            rdata=rdata,
            state_indices=state_indices,
            ax=axes[i_cond, 0],
        )
        plot_observable_trajectories(
            rdata=rdata,
            observable_indices=observable_indices,
            ax=axes[i_cond, 1],
        )
    return axes


def _time_trajectory_model_without_states(
    model: AmiciModel,
    rdatas: Union["asd.ReturnData", Sequence["asd.ReturnData"]],
    observable_ids: str | Sequence[str],
):
    """
    Visualize both, states and observables.

    Helper function for time_trajectory_model.

    Parameters
    ----------
    model:
        The amici.Model from the model of interest. Used to annotate the plot.
    rdatas:
        The data to be plotted. Each entry in the Sequence corresponds to a
        condition.
    observable_ids:
        Ids of the observables to be plotted.

    Returns
    -------
    axes:
        `matplotlib.axes.Axes` object of the plot.
    """
    # if observable_id's is not None, get indices for these
    # the AMICI plotting functions default to all indices if `None` is
    # specified.
    observable_indices = None
    if observable_ids is not None:
        observable_indices = [
            model.get_observable_ids().index(obs_id)
            for obs_id in observable_ids
        ]

    fig, axes = plt.subplots(len(rdatas))

    for i_cond, rdata in enumerate(rdatas):
        plot_observable_trajectories(
            rdata=rdata,
            observable_indices=observable_indices,
            ax=axes[i_cond] if len(rdatas) > 1 else axes,
        )
    return axes



"""Helper methods for hierarchical optimization with PEtab."""

import warnings
from typing import Literal

import pandas as pd
import petab.v1 as petab
import sympy as sp
from more_itertools import one
from petab.v1.C import (
    ESTIMATE,
    LIN,
    NOISE_PARAMETERS,
    OBSERVABLE_ID,
    OBSERVABLE_PARAMETERS,
    OBSERVABLE_TRANSFORMATION,
    PARAMETER_SEPARATOR,
)
from petab.v1.C import LOWER_BOUND as PETAB_LOWER_BOUND
from petab.v1.C import UPPER_BOUND as PETAB_UPPER_BOUND
from petab.v1.observables import get_formula_placeholders

from ..C import (
    CENSORING_BOUNDS,
    CENSORING_TYPES,
    INNER_PARAMETER_BOUNDS,
    INTERVAL_CENSORED,
    LEFT_CENSORED,
    MEASUREMENT_CATEGORY,
    MEASUREMENT_TYPE,
    ORDINAL,
    PARAMETER_TYPE,
    RELATIVE,
    RIGHT_CENSORED,
    SEMIQUANTITATIVE,
    InnerParameterType,
)
from ..C import LOWER_BOUND as PYPESTO_LOWER_BOUND
from ..C import UPPER_BOUND as PYPESTO_UPPER_BOUND


def correct_parameter_df_bounds(parameter_df: pd.DataFrame) -> pd.DataFrame:
    """Correct the bounds of inner parameters in a PEtab parameters table.

    Parameters
    ----------
    parameter_df:
        The PEtab parameters table.

    Returns
    -------
    The table, with corrected bounds.
    """

    def correct_row(row: pd.Series) -> pd.Series:
        if pd.isna(row[PARAMETER_TYPE]):
            return row
        if row[PARAMETER_TYPE] in [
            InnerParameterType.SCALING,
            InnerParameterType.OFFSET,
        ]:
            return row
        bounds = INNER_PARAMETER_BOUNDS[row[PARAMETER_TYPE]]
        row[PETAB_LOWER_BOUND] = bounds[PYPESTO_LOWER_BOUND]
        row[PETAB_UPPER_BOUND] = bounds[PYPESTO_UPPER_BOUND]
        return row

    return parameter_df.apply(correct_row, axis=1)


def validate_hierarchical_petab_problem(petab_problem: petab.Problem) -> None:
    """Validate a PEtab problem for hierarchical optimization.

    Parameters
    ----------
    petab_problem:
        The PEtab problem.
    """
    if PARAMETER_TYPE in petab_problem.parameter_df:
        # ensure we only have linear parameter scale
        inner_parameter_table = petab_problem.parameter_df[
            petab_problem.parameter_df[PARAMETER_TYPE].isin(
                [
                    InnerParameterType.OFFSET,
                    InnerParameterType.SIGMA,
                    InnerParameterType.SCALING,
                ]
            )
        ]
        if (
            petab.PARAMETER_SCALE in inner_parameter_table
            and not (
                inner_parameter_table[petab.PARAMETER_SCALE].isna()
                | (inner_parameter_table[petab.PARAMETER_SCALE] == petab.LIN)
                | (
                    inner_parameter_table[PARAMETER_TYPE]
                    != InnerParameterType.SIGMA
                )
            ).all()
        ):
            sub_df = inner_parameter_table.loc[
                :, [PARAMETER_TYPE, petab.PARAMETER_SCALE]
            ]
            raise NotImplementedError(
                "LOG and LOG10 parameter scale of inner parameters is not supported "
                "for sigma parameters. Inner parameter table:\n"
                f"{sub_df}"
            )
        elif (
            petab.PARAMETER_SCALE in inner_parameter_table
            and not (
                inner_parameter_table[petab.PARAMETER_SCALE].isna()
                | (inner_parameter_table[petab.PARAMETER_SCALE] == petab.LIN)
            ).all()
        ):
            sub_df = inner_parameter_table.loc[
                :, [PARAMETER_TYPE, petab.PARAMETER_SCALE]
            ]
            warnings.warn(
                f"LOG and LOG10 parameter scale of inner parameters is used only "
                f"for their visualization, and does not affect their optimization. "
                f"Inner parameter table:\n{sub_df}",
                stacklevel=1,
            )

        inner_parameter_df = validate_measurement_formulae(
            petab_problem=petab_problem
        )

        validate_inner_parameter_pairings(
            inner_parameter_df=inner_parameter_df
        )

    if MEASUREMENT_TYPE in petab_problem.measurement_df:
        validate_observable_data_types(petab_problem=petab_problem)


def validate_inner_parameter_pairings(
    inner_parameter_df: pd.DataFrame,
):
    """Validate the pairings of inner parameters.

    Parameters
    ----------
    inner_parameter_df:
        Each row has the offset, scaling and sigma of each measurement in
        measurement table of the PEtab problem.
    """
    scalings_with_sigmas = {}
    offsets_with_sigmas = {}

    scalings_with_offsets = {}
    offsets_with_scalings = {}

    for row in inner_parameter_df.itertuples():
        offset = getattr(row, InnerParameterType.OFFSET)
        scaling = getattr(row, InnerParameterType.SCALING)
        sigma = getattr(row, InnerParameterType.SIGMA)

        # Ensures each scaling is only ever paired with one sigma.
        if scaling is not None and sigma is not None:
            expected_sigma = scalings_with_sigmas.get(scaling)
            if expected_sigma is not None and sigma != expected_sigma:
                raise ValueError(
                    "If a scaling parameter is paired with a sigma "
                    "parameter for one measurement, all measurements with "
                    "the same scaling parameter must not have a different "
                    "sigma inner parameter. "
                    f"Scaling: `{scaling}`. Sigma: `{sigma}`. "
                    f"Expected sigma: `{expected_sigma}`."
                )
            else:
                scalings_with_sigmas[scaling] = sigma

        # Ensures each offset is only ever paired with one sigma.
        if offset is not None and sigma is not None:
            expected_sigma = offsets_with_sigmas.get(offset)
            if expected_sigma is not None and sigma != expected_sigma:
                raise ValueError(
                    "If an offset parameter is paired with a sigma "
                    "parameter for one measurement, all measurements with "
                    "the same offset parameter must not have a different "
                    "sigma inner parameter. "
                    f"Scaling: `{offset}`. Sigma: `{sigma}`. "
                    f"Expected sigma: `{expected_sigma}`."
                )
            else:
                offsets_with_sigmas[offset] = sigma

        # Ensures each scaling is only ever paired with one offset, and vice
        # versa.
        if scaling is not None and offset is not None:
            expected_offset = scalings_with_offsets.get(scaling)
            expected_scaling = offsets_with_scalings.get(offset)
            if (expected_offset is not None and offset != expected_offset) or (
                expected_scaling is not None and scaling != expected_scaling
            ):
                raise ValueError(
                    "If a scaling parameter is paired with an offset "
                    "parameter for one measurement, all measurements with "
                    "the same scaling or offset parameter must not have a "
                    f"different scaling or offset inner parameter. "
                    f"Scaling: `{scaling}`. "
                    f"Expected scaling: `{expected_scaling}`. "
                    f"Offset: `{offset}`. "
                    f"Expected offset: `{expected_offset}`."
                )
            else:
                scalings_with_offsets[scaling] = offset
                offsets_with_scalings[offset] = scaling

        # TODO can a scaling that is paired with an offset, exist without
        #      any scaling?
        #      Same for offset?
        #      Same for scaling+sigma and offset+sigma pairings?


def get_inner_parameters(
    petab_problem: petab.Problem,
) -> dict[str, InnerParameterType]:
    """Get information about the inner parameters.

    Parameters
    ----------
    petab_problem:
        The PEtab problem, with inner parameters.

    Returns
    -------
    Parameter IDs and their inner parameter types.
    """
    # TODO exclude fixed inner parameters

    if PARAMETER_TYPE not in petab_problem.parameter_df.columns:
        raise ValueError(
            "Hierarchical optimization requires specification of inner "
            f"parameters with the `{PARAMETER_TYPE}` column in the parameters "
            "table."
        )

    inner_parameters = {}
    for parameter_id, row in petab_problem.parameter_df.iterrows():
        type_str = row[PARAMETER_TYPE]
        if pd.isna(type_str):
            continue

        try:
            inner_parameters[parameter_id] = InnerParameterType(type_str)
        except ValueError as e:
            raise ValueError(
                f"Unknown inner parameter type `{type_str}`."
            ) from e

    return inner_parameters


def validate_measurement_formulae(
    petab_problem: petab.Problem,
) -> pd.DataFrame:
    """Check whether formulae associated with a measurement are valid.

    Specifically, check whether the measurement-specific observable and noise
    formulae are valid.

    Parameters
    ----------
    petab_problem:
        The PEtab problem.

    Returns
    -------
    A dataframe containing the inner parameters for each measurement.
    """
    inner_parameters = get_inner_parameters(petab_problem=petab_problem)

    inner_parameter_sets = []

    for _, measurement in petab_problem.measurement_df.iterrows():
        offset, scaling = _validate_measurement_specific_observable_formula(
            measurement=measurement,
            petab_problem=petab_problem,
            inner_parameters=inner_parameters,
        )
        sigma = _validate_measurement_specific_noise_formula(
            measurement=measurement,
            petab_problem=petab_problem,
            inner_parameters=inner_parameters,
        )
        inner_parameter_sets.append(
            [
                str(v) if v is not None else None
                for v in [offset, scaling, sigma]
            ]
        )

    return pd.DataFrame(
        data=inner_parameter_sets,
        columns=[
            InnerParameterType.OFFSET,
            InnerParameterType.SCALING,
            InnerParameterType.SIGMA,
        ],
    )


def _validate_measurement_specific_observable_formula(
    measurement: pd.Series,
    petab_problem: petab.Problem,
    inner_parameters: dict[str, InnerParameterType],
) -> tuple[InnerParameterType, InnerParameterType]:
    """Check whether a measurement observable formula is valid.

    Parameters
    ----------
    measurement:
        A row from a PEtab measurements table.
    petab_problem:
        The PEtab problem.
    inner_parameters:
        See `get_inner_parameters`.

    Returns
    -------
    The offset and scaling parameters.
    """
    formula, formula_inner_parameters = _get_symbolic_formula_from_measurement(
        measurement=measurement,
        formula_type="observable",
        petab_problem=petab_problem,
        inner_parameters=inner_parameters,
    )

    offset = None
    scaling = None

    terms = sp.Add.make_args(formula)

    for (
        formula_inner_parameter,
        inner_parameter_type,
    ) in formula_inner_parameters.items():
        if inner_parameter_type == InnerParameterType.OFFSET:
            offset = formula_inner_parameter
            appearances = [
                term for term in terms if offset in term.free_symbols
            ]
            try:
                # `one` ensures the offset occurs in exactly one term
                # `==`  ensures that the offset appearances as its own term,
                #       e.g. without being multiplied by anything
                if offset != one(appearances):
                    raise ValueError
            except Exception as e:
                raise ValueError(
                    "An offset is in the observable formula, but is not of "
                    "the expected form. The observable formula should take "
                    "the form `y = ... + offset`. Observable formula: "
                    f"`{formula}`. Offset: `{offset}`."
                ) from e
        elif inner_parameter_type == InnerParameterType.SCALING:
            scaling = formula_inner_parameter
            appearances = [
                term for term in terms if scaling in term.free_symbols
            ]
            # Should occur in exactly one term, and should be multiplied by
            # everything else in the term.
            try:
                # `one` ensures the scaling occurs in exactly one term
                factors = sp.Mul.make_args(one(appearances))
                factor_appearances = [
                    factor
                    for factor in factors
                    if scaling in factor.free_symbols
                ]
                # as for offset, ensure scaling is exactly one factor
                if scaling != one(factor_appearances):
                    raise ValueError
            except Exception as e:
                raise ValueError(
                    "A scaling is in the observable formula, but is not of "
                    "the expected form. The observable formula should take "
                    "the form `y = scaling*(...) [+ offset]`. Observable "
                    f"formula: `{formula}`. Offset: `{scaling}`."
                ) from e
        else:
            raise ValueError("Unknown error: unexpected inner parameter type.")

    return offset, scaling


def _validate_measurement_specific_noise_formula(
    measurement: pd.Series,
    petab_problem: petab.Problem,
    inner_parameters: dict[str, InnerParameterType],
) -> tuple[InnerParameterType, InnerParameterType]:
    """Check whether a measurement noise formula is valid.

    Parameters
    ----------
    measurement:
        A row from a PEtab measurements table.
    petab_problem:
        The PEtab problem.
    inner_parameters:
        See `get_inner_parameters`.

    Returns
    -------
    The sigma parameter.
    """
    formula, formula_inner_parameters = _get_symbolic_formula_from_measurement(
        measurement=measurement,
        formula_type="noise",
        petab_problem=petab_problem,
        inner_parameters=inner_parameters,
    )

    sigma = None

    for (
        formula_inner_parameter,
        inner_parameter_type,
    ) in formula_inner_parameters.items():
        if inner_parameter_type == InnerParameterType.SIGMA:
            sigma = formula_inner_parameter
            try:
                # The sigma must describe the noise completely, i.e. constitute
                # the full noise formula.
                if sigma != formula:
                    raise ValueError
            except Exception as e:
                raise ValueError(
                    "A sigma parameter does not constitute the full noise "
                    f"formula. Noise formula: `{formula}`. Sigma: `{sigma}`."
                ) from e
        else:
            raise ValueError("Unknown error: unexpected inner parameter type.")

    return sigma


def _get_symbolic_formula_from_measurement(
    measurement: pd.Series,
    formula_type: Literal["observable", "noise"],
    petab_problem: petab.Problem,
    inner_parameters: dict[str, InnerParameterType],
) -> tuple[sp.Expr, dict[sp.Symbol, InnerParameterType]]:
    """Get a symbolic representation of a formula, with overrides overridden.

    Also performs some checks to ensure only valid numbers and types of inner
    parameters are in the formula.

    Parameters
    ----------
    measurement:
        A row from the PEtab measurements table.
    formula_type:
        The formula to extract from the measurement.
    petab_problem:
        The PEtab problem.
    inner_parameters:
        See `get_inner_parameters`.

    Returns
    -------
    The symbolic formula, and its inner parameters.
    """
    observable_id = measurement[OBSERVABLE_ID]

    formula_string = petab_problem.observable_df.loc[
        observable_id, formula_type + "Formula"
    ]
    symbolic_formula = sp.sympify(formula_string)

    formula_placeholders = get_formula_placeholders(
        formula_string=formula_string,
        observable_id=observable_id,
        override_type=formula_type,
    )
    # Search for placeholders in the formula that are not allowed to be
    # overridden by inner parameters -- noise inner parameters should not
    # be in observable formulas, and vice versa.
    disallowed_formula_type = (
        "noise" if formula_type == "observable" else "observable"
    )
    disallowed_formula_placeholders = get_formula_placeholders(
        formula_string=formula_string,
        observable_id=observable_id,
        override_type=disallowed_formula_type,
    )

    if formula_placeholders:
        overrides = measurement[formula_type + "Parameters"]
        overrides = (
            overrides.split(PARAMETER_SEPARATOR)
            if isinstance(overrides, str)
            else [overrides]
        )
        subs = dict(zip(formula_placeholders, overrides, strict=True))
        symbolic_formula = symbolic_formula.subs(subs)

    if disallowed_formula_placeholders:
        disallowed_overrides = measurement[
            disallowed_formula_type + "Parameters"
        ]
        disallowed_overrides = (
            disallowed_overrides.split(PARAMETER_SEPARATOR)
            if isinstance(disallowed_overrides, str)
            else [disallowed_overrides]
        )
        disallowed_subs = dict(
            zip(
                disallowed_formula_placeholders,
                disallowed_overrides,
                strict=True,
            )
        )
        symbolic_formula = symbolic_formula.subs(disallowed_subs)

    symbolic_formula_inner_parameters = {
        sp.Symbol(inner_parameter_id): inner_parameter_type
        for inner_parameter_id, inner_parameter_type in inner_parameters.items()
        if sp.Symbol(inner_parameter_id) in symbolic_formula.free_symbols
    }

    if formula_type == "noise":
        max_parameters = 1
        expected_inner_parameter_types = [InnerParameterType.SIGMA]
    elif formula_type == "observable":
        max_parameters = 2
        expected_inner_parameter_types = [
            InnerParameterType.OFFSET,
            InnerParameterType.SCALING,
        ]

    if len(symbolic_formula_inner_parameters) > max_parameters:
        raise ValueError(
            "There are too many inner parameters in the formula. "
            f"Formula type: `{formula_type}`. Formula: `{symbolic_formula}`. "
            f"Inner parameters: `{symbolic_formula_inner_parameters}`."
        )

    inner_parameter_types = set(symbolic_formula_inner_parameters.values())
    for inner_parameter_type in inner_parameter_types:
        if inner_parameter_type not in expected_inner_parameter_types:
            raise ValueError(
                f"Unexpected inner parameter type in {formula_type} formula. "
                f"Inner parameter type: `{inner_parameter_type}`. "
                f"Formula: `{symbolic_formula}`. Expected inner parameter "
                f"types: `{expected_inner_parameter_types}`."
            )
    if len(inner_parameter_types) != len(symbolic_formula_inner_parameters):
        raise ValueError(
            "There are multiple inner parameters of the same type."
            f"Inner parameters: `{symbolic_formula_inner_parameters.values}`."
        )

    if symbolic_formula_inner_parameters:
        observable_transformation = petab_problem.observable_df.loc[
            observable_id
        ].get(OBSERVABLE_TRANSFORMATION)
        if (
            observable_transformation is not None
            and observable_transformation != LIN
        ):
            raise ValueError(
                "Non-linear observable transformations are not supported if "
                "the observable is associated with hierarchically-optimized "
                f"inner parameters. "
                f"Observable transformation: `{observable_transformation}`. "
                f"Measurement:\n{measurement}"
            )

    return symbolic_formula, symbolic_formula_inner_parameters


def validate_observable_data_types(petab_problem: petab.Problem) -> None:
    """Check whether the data types of observables are valid."""

    supported_data_types = [
        SEMIQUANTITATIVE,
        RELATIVE,
        ORDINAL,
    ] + CENSORING_TYPES

    # Get observables across data types
    observables_by_data_type = {}
    meas_df = petab_problem.measurement_df
    if MEASUREMENT_TYPE in meas_df.columns:
        petab_data_types = meas_df[meas_df[MEASUREMENT_TYPE].notna()][
            MEASUREMENT_TYPE
        ].unique()
        for data_type in petab_data_types:
            observables_by_data_type[data_type] = set(
                meas_df.loc[
                    meas_df[MEASUREMENT_TYPE] == data_type, OBSERVABLE_ID
                ]
            )
        # Check whether all data types are supported
        if not set(petab_data_types).issubset(supported_data_types):
            raise ValueError(
                f"Unsupported data type(s): `{set(petab_data_types) - set(supported_data_types)}`."
            )

    # For relative data, we search for observable parameters in the parameter table
    # and their corresponding observables in the measurement table.
    relative_observables = set()
    if PARAMETER_TYPE in petab_problem.parameter_df.columns:
        par_df = petab_problem.parameter_df[
            petab_problem.parameter_df[PARAMETER_TYPE].notna()
        ]
        meas_df_w_obs_pars = meas_df[meas_df[OBSERVABLE_PARAMETERS].notna()]
        for par_id, row in par_df.iterrows():
            if not row[ESTIMATE]:
                continue
            if row[PARAMETER_TYPE] in [
                InnerParameterType.SCALING,
                InnerParameterType.OFFSET,
            ]:
                for _, row in meas_df_w_obs_pars.iterrows():
                    if par_id in row[OBSERVABLE_PARAMETERS].split(
                        PARAMETER_SEPARATOR
                    ):
                        relative_observables.add(row[OBSERVABLE_ID])

            elif row[PARAMETER_TYPE] == InnerParameterType.SIGMA:
                corresponding_obs = set(
                    meas_df[meas_df[NOISE_PARAMETERS] == par_id][OBSERVABLE_ID]
                )
                # If a sigma parameter belongs to a semi-quantiative
                # observable, it is not a relative inner parameter.
                if SEMIQUANTITATIVE in observables_by_data_type and (
                    corresponding_obs
                    & observables_by_data_type[SEMIQUANTITATIVE]
                ):
                    continue
                relative_observables.update(corresponding_obs)

    observables_by_data_type[RELATIVE] = relative_observables

    # Check whether there is any overlap across data types
    for data_type, observables in observables_by_data_type.items():
        for (
            other_data_type,
            other_observables,
        ) in observables_by_data_type.items():
            if data_type == other_data_type or (
                data_type in CENSORING_TYPES
                and other_data_type in CENSORING_TYPES
            ):
                continue
            if observables & other_observables:
                raise ValueError(
                    "The same observable is associated with multiple data "
                    f"types. Observable(s): `{observables & other_observables}`. "
                    f"Data types: `{data_type}`, `{other_data_type}`."
                )

    # Validate ordinal measurement specification
    if ORDINAL in observables_by_data_type:
        meas_df_w_ordinals = meas_df[meas_df[MEASUREMENT_TYPE] == ORDINAL]
        if MEASUREMENT_CATEGORY not in meas_df_w_ordinals.columns:
            raise ValueError(
                "Measurement category must be specified for ordinal "
                "measurements."
            )
        for _, row in meas_df_w_ordinals.iterrows():
            try:
                category = float(row[MEASUREMENT_CATEGORY])
                if category != int(category):
                    raise ValueError
            except ValueError as e:
                raise ValueError(
                    "Measurement category for ordinal measurement must be "
                    "an integer."
                ) from e

    # Validate censored measurement specification
    if set(CENSORING_TYPES) & set(observables_by_data_type.keys()):
        meas_df_w_censored = meas_df[
            meas_df[MEASUREMENT_TYPE].isin(CENSORING_TYPES)
        ]
        if CENSORING_BOUNDS not in meas_df_w_censored.columns:
            raise ValueError(
                "Censoring bounds must be specified for censored measurements."
            )
        for _, row in meas_df_w_censored.iterrows():
            if (
                row[MEASUREMENT_TYPE] == LEFT_CENSORED
                or row[MEASUREMENT_TYPE] == RIGHT_CENSORED
            ):
                try:
                    float(row[CENSORING_BOUNDS])
                except ValueError as e:
                    raise ValueError(
                        f"Censoring bound(s) for a {row[MEASUREMENT_TYPE]}"
                        " measurement must be of type float. Failure at "
                        f"bound: `{row[CENSORING_BOUNDS]}`."
                    ) from e
            elif row[MEASUREMENT_TYPE] == INTERVAL_CENSORED:
                bounds = row[CENSORING_BOUNDS].split(PARAMETER_SEPARATOR)
                if len(bounds) != 2:
                    raise ValueError(
                        f"Censoring bounds for a {INTERVAL_CENSORED} measurement"
                        f" must be two floats separated by the separator {PARAMETER_SEPARATOR}."
                        f" Found {len(bounds)} bounds: `{bounds}`."
                    )
                try:
                    float(bounds[0])
                    float(bounds[1])
                except ValueError as e:
                    raise ValueError(
                        f"Censoring bound(s) for a {row[MEASUREMENT_TYPE]}"
                        " measurement must be of type float. Failure at "
                        f"bounds: `{bounds}`."
                    ) from e
                if float(bounds[0]) > float(bounds[1]):
                    raise ValueError(
                        f"Censoring bounds for a {INTERVAL_CENSORED}"
                        " measurement must be increasing. Failure at "
                        f"bounds: `{bounds}`."
                    )



"""Inner optimization problem in hierarchical optimization."""

from __future__ import annotations

import copy
import logging

import numpy as np
import pandas as pd

from ..C import LIN, LOG, LOG10
from .base_parameter import InnerParameter

try:
    import amici.sim.sundials as asd
    import petab.v1 as petab
    from petab.v1.C import OBSERVABLE_ID, TIME
except ImportError:
    pass

logger = logging.getLogger(__name__)


class InnerProblem:
    """
    Inner optimization problem in hierarchical optimization.

    Attributes
    ----------
    xs:
        Mapping of (inner) parameter ID to ``InnerParameters``.
    data:
        Measurement data. One matrix (`num_timepoints` x `num_observables`)
        per simulation condition. Missing observations as NaN.
    """

    def __init__(self, xs: list[InnerParameter], data: list[np.ndarray]):
        self.xs: dict[str, InnerParameter] = {
            x.inner_parameter_id: x for x in xs
        }
        self.data = copy.deepcopy(data)

        # create the joint mask of all inner problem parameters
        self.data_mask = [
            np.zeros_like(cond_data, dtype=bool) for cond_data in data
        ]
        for x in xs:
            for condition_ix, cond_mask in enumerate(self.data_mask):
                cond_mask[x.ixs[condition_ix]] = True

        # mask the data: a inner problem is aware of only the data it uses
        for i in range(len(self.data)):
            self.data[i][~self.data_mask[i]] = np.nan

        logger.debug(f"Created InnerProblem with ids {self.get_x_ids()}")

        if self.is_empty():
            raise ValueError(
                "There are no parameters in the inner problem of hierarchical "
                "optimization."
            )

    @staticmethod
    def from_petab_amici(
        petab_problem: petab.Problem,
        amici_model: asd.Model,
        edatas: list[asd.ExpData],
    ) -> InnerProblem:
        """Create an InnerProblem from a PEtab problem and AMICI objects."""

    def get_x_ids(self) -> list[str]:
        """Get IDs of inner parameters."""
        return list(self.xs.keys())

    def get_interpretable_x_ids(self) -> list[str]:
        """Get IDs of interpretable inner parameters.

        Interpretable parameters need to be easily interpretable by the user.
        Examples are scaling factors, offsets, or noise parameters. An example
        of non-interpretable inner parameters is the spline heights of spline
        approximation for semiquantitative data. It is challenging to interpret
        the meaning of these parameters based solely on their value.
        """
        return list(self.xs.keys())

    def get_interpretable_x_scales(self) -> list[str]:
        """Get scales of interpretable inner parameters."""
        return [x.scale for x in self.xs.values()]

    def get_xs_for_type(
        self, inner_parameter_type: str
    ) -> list[InnerParameter]:
        """Get inner parameters of the given type."""
        return [
            x
            for x in self.xs.values()
            if x.inner_parameter_type == inner_parameter_type
        ]

    def get_dummy_values(self, scaled: bool) -> dict[str, float]:
        """
        Get dummy parameter values.

        Get parameter values to be used for simulation before their optimal
        values are computed.

        Parameters
        ----------
        scaled:
            Whether the parameters should be returned on parameter scale (``True``)
            or on linear scale (``False``).
        """
        return {
            x.inner_parameter_id: (
                scale_value(x.dummy_value, x.scale)
                if scaled
                else x.dummy_value
            )
            for x in self.xs.values()
        }

    def get_for_id(self, inner_parameter_id: str) -> InnerParameter:
        """Get InnerParameter for the given parameter ID."""
        try:
            return self.xs[inner_parameter_id]
        except KeyError:
            raise KeyError(
                f"Cannot find parameter with id {inner_parameter_id}."
            ) from None

    def is_empty(self) -> bool:
        """Check for emptiness.

        Returns
        -------
        ``True`` if there aren't any parameters associated with this problem,
        ``False`` otherwise.
        """
        return len(self.xs) == 0

    def get_bounds(self) -> tuple[np.ndarray, np.ndarray]:
        """Get bounds of inner parameters."""
        lb = np.asarray([x.lb for x in self.xs.values()])
        ub = np.asarray([x.ub for x in self.xs.values()])
        return lb, ub

    def get_interpretable_x_bounds(self) -> tuple[np.ndarray, np.ndarray]:
        """Get bounds of interpretable inner parameters."""
        interpretable_x_ids = self.get_interpretable_x_ids()
        lb = np.asarray(
            [
                x.lb
                for x in self.xs.values()
                if x.inner_parameter_id in interpretable_x_ids
            ]
        )
        ub = np.asarray(
            [
                x.ub
                for x in self.xs.values()
                if x.inner_parameter_id in interpretable_x_ids
            ]
        )
        return lb, ub


class AmiciInnerProblem(InnerProblem):
    """
    Inner optimization problem in hierarchical optimization.

    For use with AMICI objects.

    Attributes
    ----------
    edatas:
        AMICI ``ExpDataView``s for each simulation condition.
    """

    def __init__(self, edatas: list[asd.ExpData], **kwargs):
        super().__init__(**kwargs)

    def check_edatas(self, edatas: list[asd.ExpData]) -> bool:
        """Check for consistency in data.

        Currently only checks for the actual data values. e.g., timepoints are
        not compared.

        Parameters
        ----------
        edatas:
            A data set. Will be checked against the data set provided to the
            constructor.

        Returns
        -------
        Whether the data sets are consistent.
        """
        # TODO replace but edata1==edata2 once this makes it into amici
        #  https://github.com/AMICI-dev/AMICI/issues/1880
        data = [asd.ExpDataView(edata)["measurements"] for edata in edatas]

        # Mask the data using the inner problem mask. This is necessary
        # because the inner problem is aware of only the data it uses.
        for i in range(len(data)):
            data[i][~self.data_mask[i]] = np.nan

        if len(self.data) != len(data):
            return False

        for data0, data1 in zip(self.data, data, strict=True):
            if not np.array_equal(data0, data1, equal_nan=True):
                return False

        return True


def scale_value_dict(
    dct: dict[str, float], problem: InnerProblem
) -> dict[str, float]:
    """Scale a value dictionary."""
    scaled_dct = {}
    for key, val in dct.items():
        x = problem.get_for_id(key)
        scaled_dct[key] = scale_value(val, x.scale)
    return scaled_dct


def scale_value(val: float | np.array, scale: str) -> float | np.array:
    """Scale a single value."""
    if scale == LIN:
        return val
    if scale == LOG:
        return np.log(val)
    if scale == LOG10:
        return np.log10(val)
    raise ValueError(f"Scale {scale} not recognized.")


def scale_back_value_dict(
    dct: dict[str, float], problem: InnerProblem
) -> dict[str, float]:
    """Scale back a value dictionary."""
    scaled_dct = {}
    for key, val in dct.items():
        x = problem.get_for_id(key)
        scaled_dct[key] = scale_back_value(val, x.scale)
    return scaled_dct


def scale_back_value(val: float | np.array, scale: str) -> float | np.array:
    """Scale back a single value."""
    if scale == LIN:
        return val
    if scale == LOG:
        return np.exp(val)
    if scale == LOG10:
        return 10**val
    raise ValueError(f"Scale {scale} not recognized.")


def ix_matrices_from_arrays(
    ixs: dict[str, list[tuple[int, int, int]]], edatas: list[np.array]
) -> dict[str, list[np.array]]:
    """
    Convert mapping of parameters to measurements to matrix form.

    Returns
    -------
    A dictionary mapping parameter ID to a list of Boolean matrices, one per
    simulation condition. Therein, ``True`` indicates that the respective
    parameter is used for the model output at the respective timepoint,
    observable and condition index.
    """
    ix_matrices = {
        id: [np.zeros_like(edata, dtype=bool) for edata in edatas]
        for id in ixs
    }
    for id, arr in ixs.items():
        matrices = ix_matrices[id]
        for condition_ix, time_ix, observable_ix in arr:
            matrices[condition_ix][time_ix, observable_ix] = True
    return ix_matrices


def _get_timepoints_with_replicates(
    measurement_df: pd.DataFrame,
) -> list[float]:
    """
    Get list of timepoints including replicate measurements.

    :param measurement_df:
        PEtab measurement table subset for a single condition.

    :return:
        Sorted list of timepoints, including multiple timepoints accounting
        for replicate measurements.
    """
    # create sorted list of all timepoints for which measurements exist
    timepoints = sorted(measurement_df[TIME].unique().astype(float))

    # find replicate numbers of time points
    timepoints_w_reps = []
    for time in timepoints:
        # subselect for time
        df_for_time = measurement_df[measurement_df.time == time]
        # rep number is maximum over rep numbers for observables
        n_reps = max(df_for_time.groupby([OBSERVABLE_ID, TIME]).size())
        # append time point n_rep times
        timepoints_w_reps.extend([time] * n_reps)

    return timepoints_w_reps



import logging
from collections.abc import Callable, Iterable
from warnings import warn

from ..engine import Engine, SingleCoreEngine
from ..history import HistoryOptions
from ..problem import Problem
from ..result import Result
from ..startpoint import StartpointMethod, to_startpoint_method, uniform
from ..store import autosave
from .optimizer import Optimizer, ScipyOptimizer
from .options import OptimizeOptions
from .task import OptimizerTask
from .util import (
    assign_ids,
    bound_n_starts_from_env,
    postprocess_hdf5_history,
    preprocess_hdf5_history,
)

logger = logging.getLogger(__name__)


def minimize(
    problem: Problem,
    optimizer: Optimizer = None,
    n_starts: int = 100,
    ids: Iterable[str] = None,
    startpoint_method: StartpointMethod | Callable | bool = None,
    result: Result = None,
    engine: Engine = None,
    progress_bar: bool = None,
    options: OptimizeOptions = None,
    history_options: HistoryOptions = None,
    filename: str | Callable | None = None,
    overwrite: bool = False,
) -> Result:
    """
    Do multistart optimization.

    Parameters
    ----------
    problem:
        The problem to be solved.
    optimizer:
        The optimizer to be used `n_starts` times.
    n_starts:
        Number of starts of the optimizer.
    ids:
        Ids assigned to the startpoints.
    startpoint_method:
        Method for how to choose start points. ``False`` means the optimizer does
        not require start points, e.g. for the :class:`pypesto.optimize.PyswarmOptimizer`.
        **Deprecated. Use ``problem.startpoint_method`` instead.**
    result:
        A result object to append the optimization results to. For example,
        one might append more runs to a previous optimization. If None,
        a new object is created.
    engine:
        Parallelization engine. Defaults to sequential execution using
        :class:`pypesto.engine.SingleCoreEngine`.
    progress_bar:
        Whether to display a progress bar.
    options:
        Various options applied to the multistart optimization.
    history_options:
        Optimizer history options.
    filename:
        Name of the hdf5 file, where the result will be saved. Default is
        None, which deactivates automatic saving. If set to
        ``Auto`` it will automatically generate a file named
        ``year_month_day_profiling_result.hdf5``.
        Optionally a method, see docs for :func:`pypesto.store.auto.autosave`.
    overwrite:
        Whether to overwrite `result/optimization` in the autosave file
        if it already exists.

    Returns
    -------
    Result object containing the results of all multistarts in
    `result.optimize_result`.
    """
    # optimizer
    if optimizer is None:
        optimizer = ScipyOptimizer()

    # number of starts
    n_starts = bound_n_starts_from_env(n_starts)

    # startpoint method
    if startpoint_method is None:
        if problem.startpoint_method is None:
            startpoint_method = uniform
        else:
            startpoint_method = problem.startpoint_method
    else:
        warn(
            "Passing `startpoint_method` directly is deprecated, "
            "use `problem.startpoint_method` instead.",
            DeprecationWarning,
            stacklevel=2,
        )

    # convert startpoint method to class instance
    startpoint_method = to_startpoint_method(startpoint_method)

    # check options
    if options is None:
        options = OptimizeOptions()
    options = OptimizeOptions.assert_instance(options)

    # history options
    if history_options is None:
        history_options = HistoryOptions()
    history_options = HistoryOptions.assert_instance(history_options)

    # assign startpoints
    startpoints = startpoint_method(
        n_starts=n_starts,
        problem=problem,
    )

    ids = assign_ids(
        n_starts=n_starts,
        ids=ids,
        result=result,
    )

    # prepare result
    if result is None:
        result = Result(problem)

    # engine
    if engine is None:
        engine = SingleCoreEngine()

    # change to one hdf5 storage file per start if parallel and if hdf5
    history_file = history_options.storage_file
    history_requires_postprocessing = preprocess_hdf5_history(
        history_options, engine
    )

    # define tasks
    tasks = []
    for startpoint, id in zip(startpoints, ids, strict=True):
        task = OptimizerTask(
            optimizer=optimizer,
            problem=problem,
            x0=startpoint,
            id=id,
            history_options=history_options,
            optimize_options=options,
        )
        tasks.append(task)

    # perform multistart optimization
    ret = engine.execute(tasks, progress_bar=progress_bar)

    # merge hdf5 history files
    if history_requires_postprocessing:
        postprocess_hdf5_history(ret, history_file, history_options)

    # aggregate results
    for optimizer_result in ret:
        result.optimize_result.append(optimizer_result)

    # sort by best fval
    result.optimize_result.sort()

    # if history file provided, set storage file to that one
    if filename == "Auto" and history_file is not None:
        filename = history_file
    autosave(
        filename=filename,
        result=result,
        store_type="optimize",
        overwrite=overwrite,
    )

    return result



import logging
from collections.abc import Callable
from typing import Literal

import numpy as np

from ..C import GRAD
from ..optimize import OptimizeOptions, Optimizer
from ..problem import Problem
from ..result import OptimizerResult, ProfilerResult
from .options import ProfileOptions

logger = logging.getLogger(__name__)


def walk_along_profile(
    current_profile: ProfilerResult,
    problem: Problem,
    par_direction: Literal[1, -1],
    optimizer: Optimizer,
    options: ProfileOptions,
    create_next_guess: Callable,
    global_opt: float,
    i_par: int,
    max_tries: int = 10,
) -> ProfilerResult:
    """
    Compute half a profile.

    Walk ahead in positive direction until some stopping criterion is
    fulfilled. A two-sided profile is obtained by flipping the profile
    direction.

    Parameters
    ----------
    current_profile:
        The profile which should be computed
    problem:
        The problem to be solved.
    par_direction:
        Indicates profiling direction (+1, -1: ascending, descending)
    optimizer:
        The optimizer to be used along each profile.
    global_opt:
        log-posterior value of the global optimum
    options:
        Various options applied to the profile optimization.
    create_next_guess:
        Handle of the method which creates the next profile point proposal
    i_par:
        index for the current parameter
    max_tries:
        If optimization at a given profile point fails, retry optimization
        ``max_tries`` times with randomly sampled starting points.

    Returns
    -------
    The current profile, modified in-place.
    """
    if par_direction not in (-1, 1):
        raise AssertionError("par_direction must be -1 or 1")

    # while loop for profiling (will be exited by break command)
    while True:
        # get current position on the profile path
        x_now = current_profile.x_path[:, -1]
        color_now = current_profile.color_path[-1]

        # check if the next profile point needs to be computed
        # ... check bounds
        if par_direction == -1 and x_now[i_par] <= problem.lb_full[[i_par]]:
            break
        if par_direction == 1 and x_now[i_par] >= problem.ub_full[[i_par]]:
            break

        # ... check likelihood ratio
        if (
            not options.whole_path
            and current_profile.ratio_path[-1] < options.ratio_min
        ):
            break

        optimization_successful = False
        max_step_reduce_factor = 1.0

        while not optimization_successful:
            # Check max_step_size is not reduced below min_step_size
            if (
                options.max_step_size * max_step_reduce_factor
                < options.min_step_size
            ):
                logger.warning(
                    "Max step size reduced below min step size. "
                    "Setting a lower min step size can help avoid this issue."
                )
                break

            # compute the new start point for optimization
            x_next = create_next_guess(
                x_now,
                i_par,
                par_direction,
                options,
                current_profile,
                problem,
                global_opt,
                1.0,
                max_step_reduce_factor,
            )

            # fix current profiling parameter to current value and set start point
            problem.fix_parameters(i_par, x_next[i_par])
            startpoint = x_next[problem.x_free_indices]

            if startpoint.size > 0:
                optimizer_result = optimizer.minimize(
                    problem=problem,
                    x0=startpoint,
                    id=str(0),
                    optimize_options=OptimizeOptions(
                        allow_failed_starts=False
                    ),
                )

                if np.isfinite(optimizer_result.fval):
                    optimization_successful = True
                    if max_step_reduce_factor == 1.0:
                        # The color of the point is set to black if no changes were made
                        color_next = np.array([0, 0, 0, 1])
                    else:
                        # The color of the point is set to red if the max_step_size was reduced
                        color_next = np.array([1, 0, 0, 1])
                else:
                    max_step_reduce_factor *= 0.5
                    logger.warning(
                        f"Optimization at {problem.x_names[i_par]}={x_next[i_par]} failed. "
                        f"Reducing max_step_size to {options.max_step_size * max_step_reduce_factor}."
                    )
            else:
                # if too many parameters are fixed, there is nothing to do ...
                fval = problem.objective(np.array([]))
                optimizer_result = OptimizerResult(
                    id="0",
                    x=np.array([]),
                    fval=fval,
                    n_fval=0,
                    n_grad=0,
                    n_res=0,
                    n_hess=0,
                    n_sres=0,
                    x0=np.array([]),
                    fval0=fval,
                    time=0,
                )
                optimizer_result.update_to_full(problem=problem)
                optimization_successful = True
                color_next = np.concatenate((color_now[:3], [0.3]))

        if not optimization_successful:
            # Cannot optimize successfully by reducing max_step_size
            # Let's try to optimize by increasing min_step_size
            logger.warning(
                f"Failing to optimize at {problem.x_names[i_par]}={x_next[i_par]} after reducing max_step_size."
                f"Trying to increase min_step_size."
            )
            min_step_increase_factor = 1.25
        while not optimization_successful:
            # Check min_step_size is not increased above max_step_size
            if (
                options.min_step_size * min_step_increase_factor
                > options.max_step_size
            ):
                logger.warning(
                    "Min step size increased above max step size. "
                    "Setting a higher max step size can help avoid this issue."
                )
                break

            # compute the new start point for optimization
            x_next = create_next_guess(
                x_now,
                i_par,
                par_direction,
                options,
                current_profile,
                problem,
                global_opt,
                min_step_increase_factor,
                1.0,
            )

            # fix current profiling parameter to current value and set start point
            problem.fix_parameters(i_par, x_next[i_par])
            startpoint = x_next[problem.x_free_indices]

            optimizer_result = optimizer.minimize(
                problem=problem,
                x0=startpoint,
                id=str(0),
                optimize_options=OptimizeOptions(allow_failed_starts=False),
            )

            if np.isfinite(optimizer_result.fval):
                optimization_successful = True
                # The color of the point is set to blue if the min_step_size was increased
                color_next = np.array([0, 0, 1, 1])
            else:
                min_step_increase_factor *= 1.25
                logger.warning(
                    f"Optimization at {problem.x_names[i_par]}={x_next[i_par]} failed. "
                    f"Increasing min_step_size to {options.min_step_size * min_step_increase_factor}."
                )

        if not optimization_successful:
            # Cannot optimize successfully by reducing max_step_size or increasing min_step_size
            # sample a new starting point for another attempt for max_tries times
            logger.warning(
                f"Failing to optimize at {problem.x_names[i_par]}={x_next[i_par]} after reducing max_step_size."
                f"Trying to sample {max_tries} new starting points."
            )

            x_next = create_next_guess(
                x_now,
                i_par,
                par_direction,
                options,
                current_profile,
                problem,
                global_opt,
                1.0,
                1.0,
            )

            problem.fix_parameters(i_par, x_next[i_par])

            for i_optimize_attempt in range(max_tries):
                startpoint = problem.startpoint_method(
                    n_starts=1, problem=problem
                )[0]

                optimizer_result = optimizer.minimize(
                    problem=problem,
                    x0=startpoint,
                    id=str(i_optimize_attempt),
                    optimize_options=OptimizeOptions(
                        allow_failed_starts=False
                    ),
                )
                if np.isfinite(optimizer_result.fval):
                    # The color of the point is set to green if the parameter was resampled
                    color_next = np.array([0, 1, 0, 1])
                    break

                logger.warning(
                    f"Optimization at {problem.x_names[i_par]}={x_next[i_par]} failed."
                )
            else:
                raise RuntimeError(
                    f"Computing profile point failed. Could not find a finite solution after {max_tries} attempts."
                )

        logger.debug(
            f"Optimization successful for {problem.x_names[i_par]}={x_next[i_par]:.4f}. "
            f"Start fval {problem.objective(x_next[problem.x_free_indices]):.6f}, end fval {optimizer_result.fval:.6f}."
        )
        if optimizer_result[GRAD] is not None:
            gradnorm = np.linalg.norm(
                optimizer_result[GRAD][problem.x_free_indices]
            )
        else:
            gradnorm = np.nan

        current_profile.append_profile_point(
            x=optimizer_result.x,
            fval=optimizer_result.fval,
            ratio=np.exp(global_opt - optimizer_result.fval),
            gradnorm=gradnorm,
            time=optimizer_result.time,
            color=color_next,
            exitflag=optimizer_result.exitflag,
            n_fval=optimizer_result.n_fval,
            n_grad=optimizer_result.n_grad,
            n_hess=optimizer_result.n_hess,
        )

    # free the profiling parameter again
    problem.unfix_parameters(i_par)

    return current_profile



"""DynestySampler class."""

# TODO set `nlive` to > a few * n_params**2
#      - https://github.com/joshspeagle/dynesty/issues/301#issuecomment-888461184
#      - seems easy for `NestedSampler(..., live_points=...)`, and for
#        `DynamicNestedSampler(..., nlive0=...)`.
# TODO set one of the initial live points to the best vector from optimization
#      - this should hopefully ensure a good estimate for `L_max`, used for
#        stopping criteria
#      - seems easy for `NestedSampler(..., live_points)`, but not for
#        `DynamicNestedSampler`.
# TODO allow selection of bounding methods
#      - which bounding method works well for high-dimension problems?
#        - possibly overlapping balls/cubes (reduced tuning parameters), or
#          unit cube (but only if unimodal?)
#          - ellipsoids seem to be recommended
#        - requires benchmarking...
# TODO consider "inner" sampling method
#      - Hamiltonian for large problems (D > 20) with gradients
#        - need to reshape gradients to unit cube
#        - need to provide gradients, hopefully without additional obj call
#        - or use `compute_jac` option, probably expensive
#      - multivariate slice for large problems (D > 20) without gradients
# TODO consider reflective boundaries

from __future__ import annotations

import importlib
import logging
import warnings

import cloudpickle  # noqa: S403
import numpy as np

from ..C import OBJECTIVE_NEGLOGLIKE, OBJECTIVE_NEGLOGPOST
from ..problem import Problem
from ..result import McmcPtResult
from .sampler import Sampler, SamplerImportError

dynesty_pickle = cloudpickle
if importlib.util.find_spec("dynesty") is None:
    dynesty = type("", (), {})()
    dynesty.results = type("", (), {})()
    dynesty.results.Results = None
else:
    import dynesty

    dynesty.utils.pickle_module = dynesty_pickle
logger = logging.getLogger(__name__)


class DynestySampler(Sampler):
    """Use dynesty for sampling.

    NB: `get_samples` returns MCMC-like samples, by resampling original
    `dynesty` samples according to their importance weights. This is because
    the original samples contain many low-likelihood samples.
    To work with the original samples, modify the results object with
    `pypesto_result.sample_result = sampler.get_original_samples()`, where
    `sampler` is an instance of `pypesto.sample.DynestySampler`. The original
    dynesty results object is available at `sampler.raw_results`.

    NB: the dynesty samplers can be customized significantly, by providing
    `sampler_args` and `run_args` to your `pypesto.sample.DynestySampler()`
    call. For example, code to parallelize dynesty is provided in pyPESTO's
    `sampler_study.ipynb` notebook.

    Wrapper around https://dynesty.readthedocs.io/en/stable/, see there for
    details.
    """

    def __init__(
        self,
        sampler_args: dict = None,
        run_args: dict = None,
        dynamic: bool = True,
        objective_type: str = OBJECTIVE_NEGLOGPOST,
        prior_transform: callable = None,
    ):
        """
        Initialize sampler.

        Parameters
        ----------
        sampler_args:
            Further keyword arguments that are passed on to the `__init__`
            method of the dynesty sampler.
        run_args:
            Further keyword arguments that are passed on to the `run_nested`
            method of the dynesty sampler.
        dynamic:
            Whether to use dynamic or static nested sampling.
        objective_type:
            The objective to optimize (as defined in `pypesto.problem`). Either
            `pypesto.C.OBJECTIVE_NEGLOGLIKE` or
            `pypesto.C.OBJECTIVE_NEGLOGPOST`. If
            `pypesto.C.OBJECTIVE_NEGLOGPOST`, then `x_priors` have to
            be defined in the problem.
        prior_transform:
            A function converting a sample from the unit cube to actual prior. If not provided, the default
            `prior_transform` function is used, which assumes uniform priors.
        """
        if importlib.util.find_spec("dynesty") is None:
            raise SamplerImportError("dynesty")

        super().__init__()

        self.dynamic = dynamic

        if sampler_args is None:
            sampler_args = {}
        self.sampler_args: dict = sampler_args

        if run_args is None:
            run_args = {}
        self.run_args: dict = run_args

        if objective_type not in [OBJECTIVE_NEGLOGPOST, OBJECTIVE_NEGLOGLIKE]:
            raise ValueError(
                f"Objective has to be either '{OBJECTIVE_NEGLOGPOST}' or '{OBJECTIVE_NEGLOGLIKE}' "
                f"as defined in pypesto.problem."
            )
        self.objective_type = objective_type

        if prior_transform is None:
            # if priors are uniform, we can use the default prior transform (assuming that bounds are set correctly)
            logger.warning(
                "Assuming 'prior_transform' is correctly specified. If 'x_priors' is not uniform, 'prior_transform'"
                " has to be adjusted accordingly."
            )
            self.prior_transform = self.prior_transform_from_uniform
        else:
            self.prior_transform = prior_transform

        # set in initialize
        self.problem: Problem | None = None
        self.sampler: (
            dynesty.DynamicNestedSampler | dynesty.NestedSampler | None
        ) = None

    def loglikelihood(self, x):
        """Log-probability density function."""
        # check if parameter lies within bounds
        if any(x < self.problem.lb) or any(x > self.problem.ub):
            return -np.inf
        # invert sign
        if self.objective_type == OBJECTIVE_NEGLOGPOST:
            # problem.objective returns negative log-posterior
            # compute log-likelihood by subtracting log-prior
            return -1.0 * (
                self.problem.objective(x) - self.problem.x_priors(x)
            )
        # problem.objective returns negative log-likelihood
        return -1.0 * self.problem.objective(x)

    def prior_transform_from_uniform(
        self, prior_sample: np.ndarray
    ) -> np.ndarray:
        """Transform prior sample from unit cube to pyPESTO prior.

        Parameters
        ----------
        prior_sample:
            The prior sample, provided by dynesty.

        Returns
        -------
        The transformed prior sample.
        """
        return (
            prior_sample * (self.problem.ub - self.problem.lb)
            + self.problem.lb
        )

    def initialize(
        self,
        problem: Problem,
        x0: np.ndarray | list[np.ndarray] = None,
    ) -> None:
        """Initialize the sampler."""
        self.problem = problem

        sampler_class = dynesty.NestedSampler
        if self.dynamic:
            sampler_class = dynesty.DynamicNestedSampler

        # check if objective fits to the pyPESTO problem
        if self.objective_type == OBJECTIVE_NEGLOGPOST:
            if self.problem.x_priors is None:
                # objective is the negative log-posterior, but no priors are defined
                # sampler needs the likelihood
                raise ValueError(
                    f"x_priors have to be defined in the problem if objective is '{OBJECTIVE_NEGLOGPOST}'."
                )
        else:
            # if objective is the negative log likelihood, we will ignore x_priors even if they are defined
            if self.problem.x_priors is not None:
                logger.warning(
                    f"Assuming '{OBJECTIVE_NEGLOGLIKE}' as objective. "
                    f"'x_priors' defined in the problem will be ignored."
                )

        # initialize sampler
        self.sampler = sampler_class(
            loglikelihood=self.loglikelihood,
            prior_transform=self.prior_transform,
            ndim=len(self.problem.x_free_indices),
            **self.sampler_args,
        )

        # TODO somehow use optimized vector as one of the initial live points

    def sample(self, n_samples: int, beta: float = None) -> None:
        """Return the most recent sample state."""
        if n_samples is not None:
            logger.warning(
                "`n_samples` was specified but this is incompatible with "
                "`dynesty` samplers, as they run until other stopping "
                "criteria are satisfied."
            )
        if beta is not None:
            logger.warning(
                "The temperature of a `dynesty` sampler was set, but this is "
                "irrelevant for `dynesty` samplers."
            )

        self.sampler.run_nested(**self.run_args)

    @property
    def results(self):
        """Deprecated in favor of `raw_results`."""
        warnings.warn(
            "Accessing dynesty results via `sampler.results` is "
            "deprecated. Please use `sampler.raw_results` instead.",
            DeprecationWarning,
            stacklevel=1,
        )
        return self.raw_results

    @property
    def raw_results(self):
        """Get the raw dynesty results."""
        return self.sampler.results

    def save_internal_sampler(self, filename: str) -> None:
        """Save the state of the internal dynesty sampler.

        This makes it easier to analyze the original dynesty samples, after
        sampling, with `restore_internal`.

        Parameters
        ----------
        filename:
            The internal sampler will be saved here.
        """
        dynesty.utils.save_sampler(
            sampler=self.sampler,
            fname=filename,
        )

    def restore_internal_sampler(self, filename: str) -> None:
        """Restore the state of the internal dynesty sampler.

        Parameters
        ----------
        filename:
            The internal sampler will be saved here.
        """
        pool = self.sampler_args.get("pool", None)
        self.sampler = dynesty.utils.restore_sampler(
            fname=filename,
            pool=pool,
        )

    def get_original_samples(self) -> McmcPtResult:
        """Get the samples into the fitting pypesto format.

        Returns
        -------
        The pyPESTO sample result.
        """
        return get_original_dynesty_samples(sampler=self)

    def get_samples(self) -> McmcPtResult:
        """Get MCMC-like samples into the fitting pypesto format.

        Returns
        -------
        The pyPESTO sample result.
        """
        return get_mcmc_like_dynesty_samples(sampler=self)


def _get_raw_results(
    sampler: DynestySampler,
    raw_results: dynesty.result.Results,
) -> dynesty.results.Results:
    if (sampler is None) == (raw_results is None):
        raise ValueError(
            "Please supply exactly one of `sampler` or `raw_results`."
        )

    if raw_results is not None:
        return raw_results

    if not isinstance(sampler, DynestySampler):
        raise ValueError(
            "Please provide a pyPESTO `DynestySampler` if using "
            "the `sampler` argument of this method."
        )

    return sampler.raw_results


def save_raw_results(sampler: DynestySampler, filename: str) -> None:
    """Save dynesty sampler results to file.

    Restoring the dynesty sampler on a different computer than the one that
    samples were generated is problematic (e.g. an AMICI model might get
    compiled automatically). This method should avoid that, by only saving
    the results.

    Parameters
    ----------
    sampler:
        The pyPESTO `DynestySampler` object used during sampling.
    filename:
        The file where the results will be saved.
    """
    raw_results = _get_raw_results(sampler=sampler, raw_results=None)
    with open(filename, "wb") as f:
        dynesty_pickle.dump(raw_results, f)


def load_raw_results(filename: str) -> dynesty.results.Results:
    """Load dynesty sample results from file.

    Parameters
    ----------
    filename:
        The file where the results will be loaded from.
    """
    with open(filename, "rb") as f:
        raw_results = dynesty_pickle.load(f)
    return raw_results


def get_original_dynesty_samples(
    sampler: DynestySampler = None,
    raw_results: dynesty.results.Results = None,
) -> McmcPtResult:
    """Get original dynesty samples.

    Only one of `sampler` or `raw_results` should be provided.

    Parameters
    ----------
    sampler:
        The pyPESTO `DynestySampler` object with sampling results.
    raw_results:
        The raw results. See :func:`save_raw_results` and
        :func:`load_raw_results`.

    Returns
    -------
    The sample result.
    """
    raw_results = _get_raw_results(sampler=sampler, raw_results=raw_results)

    trace_x = np.array([raw_results.samples])
    trace_neglogpost = -np.array([raw_results.logl])

    # the sampler uses custom adaptive priors
    trace_neglogprior = np.full(trace_neglogpost.shape, np.nan)
    # the sampler uses temperature 1
    betas = np.array([1.0])

    result = McmcPtResult(
        trace_x=trace_x,
        trace_neglogpost=trace_neglogpost,
        trace_neglogprior=trace_neglogprior,
        betas=betas,
    )

    return result


def get_mcmc_like_dynesty_samples(
    sampler: DynestySampler = None,
    raw_results: dynesty.results.Results = None,
) -> McmcPtResult:
    """Get MCMC-like samples.

    Only one of `sampler` or `raw_results` should be provided.

    Parameters
    ----------
    sampler:
        The pyPESTO `DynestySampler` object with sampling results.
    raw_results:
        The raw results. See :func:`save_raw_results` and
        :func:`load_raw_results`.

    Returns
    -------
    The sample result.
    """
    raw_results = _get_raw_results(sampler=sampler, raw_results=raw_results)

    if len(raw_results.importance_weights().shape) != 1:
        raise ValueError(
            "Unknown error. The dynesty importance weights are not a 1D array."
        )
    # resample according to importance weights
    indices = dynesty.utils.resample_equal(
        np.arange(raw_results.importance_weights().shape[0]),
        raw_results.importance_weights(),
    )

    trace_x = np.array([raw_results.samples[indices]])
    trace_neglogpost = -np.array([raw_results.logl[indices]])

    trace_neglogprior = np.array([np.full((len(indices),), np.nan)])
    betas = np.array([1.0])

    result = McmcPtResult(
        trace_x=trace_x,
        trace_neglogpost=trace_neglogpost,
        trace_neglogprior=trace_neglogprior,
        betas=betas,
    )
    return result



"""Objective function for PEtab models using the PEtab simulator."""

from __future__ import annotations

try:
    import petab.v1 as petab
    from petab.v1.simulate import Simulator as PetabSimulator
except ImportError:
    petab = None
from collections import OrderedDict
from collections.abc import Sequence

import numpy as np

from ..C import FVAL, MODE_FUN, MODE_RES, RES, ModeType
from .base import ObjectiveBase, ResultDict


class PetabSimulatorObjective(ObjectiveBase):
    """Objective function for PEtab models using the PEtab simulator."""

    def __init__(
        self,
        simulator: PetabSimulator,
        x_names: Sequence[str] | None = None,
    ):
        """Initialize the PEtab simulator objective function.

        Parameters
        ----------
        petab_problem:
            The PEtab problem.
        simulator:
            The PEtab simulator.
        x_names:
            Names of optimization parameters.
        """
        if petab is None:
            raise ImportError(
                "The `petab` package is required for this objective function."
            )
        self.simulator = simulator
        self.petab_problem = self.simulator.petab_problem
        if x_names is None:
            x_names = list(self.petab_problem.get_x_ids())
        super().__init__(x_names=x_names)

    def replace_parameters(self, x: np.ndarray):
        """Replace the parameters in the PEtab problem with the given values.

        Parameters
        ----------
        x:
            Parameter vector for optimization.
        """
        x_dict = OrderedDict(zip(self._x_names, x, strict=True))
        x_unscaled = self.petab_problem.unscale_parameters(x_dict)
        par_df = self.petab_problem.parameter_df
        par_df["nominalValue"] = par_df.index.map(x_unscaled)
        self.simulator.set_parameters(x_unscaled)

    def call_unprocessed(
        self,
        x: np.ndarray,
        sensi_orders: tuple[int, ...],
        mode: ModeType,
        return_dict: bool,
        **kwargs,
    ) -> ResultDict:
        """See :meth:`ObjectiveBase.call_unprocessed`."""

        self.replace_parameters(x)
        sim_df = self.simulator.simulate(noise=False, as_measurement=False)
        result = {}
        result["simulations"] = sim_df
        if mode == MODE_FUN:
            result[FVAL] = -petab.calculate_llh(
                measurement_dfs=self.petab_problem.measurement_df,
                simulation_dfs=sim_df,
                observable_dfs=self.petab_problem.observable_df,
                parameter_dfs=self.petab_problem.parameter_df,
            )
        elif mode == MODE_RES:
            result[RES] = petab.calculate_residuals(
                measurement_dfs=self.petab_problem.measurement_df,
                simulation_dfs=sim_df,
                observable_dfs=self.petab_problem.observable_df,
                parameter_dfs=self.petab_problem.parameter_df,
            )
        return result

    def check_sensi_orders(
        self,
        sensi_orders: tuple[int, ...],
        mode: ModeType,
    ) -> bool:
        """See :class:`ObjectiveBase` documentation."""
        if not sensi_orders:
            return True
        sensi_order = max(sensi_orders)
        max_sensi_order = 0

        return sensi_order <= max_sensi_order



import numpy as np

from .base_problem import InnerProblem


class InnerSolver:
    """Solver for an inner optimization problem."""

    def initialize(self):
        """
        (Re-)initialize the solver.

        Default: Do nothing.
        """

    def solve(
        self,
        problem: InnerProblem,
        sim: list[np.ndarray],
        sigma: list[np.ndarray],
        scaled: bool,
    ) -> dict[str, float] | list:
        """Solve the subproblem.

        Parameters
        ----------
        problem:
            The inner problem to solve.
        sim:
            List of model output matrices, as provided in AMICI's
            ``ReturnData.y``. Same order as simulations in the
            PEtab problem.
        sigma:
            List of sigma matrices from the model, as provided in AMICI's
            ``ReturnData.sigmay``. Same order as simulations in the
            PEtab problem.
        scaled:
            Whether to scale the results to the parameter scale specified in
            ``problem``.

        Returns
        -------
        A dictionary of inner parameter ids and their optimal values for the
        relative inner problem, or a list of inner optimization results
        for the semiquantitative and ordinal inner problems.
        """



"""Functionality related to using a PEtab Select model selection method."""

import logging
from collections.abc import Callable
from dataclasses import dataclass
from enum import StrEnum
from typing import Any

import numpy as np
import petab_select
from petab_select import (
    CANDIDATE_SPACE,
    MODELS,
    PREDECESSOR_MODEL,
    UNCALIBRATED_MODELS,
    VIRTUAL_INITIAL_MODEL,
    CandidateSpace,
    Criterion,
    Method,
    Model,
    Models,
)

from ..problem import Problem
from .model_problem import TYPE_POSTPROCESSOR, ModelProblem


class MethodSignalProceed(StrEnum):
    """Indicators for how a model selection method should proceed."""

    # TODO move to PEtab Select?
    STOP = "stop"
    CONTINUE = "continue"


@dataclass
class MethodSignal:
    """The state of a model selection method after a single model calibration.

    Attributes
    ----------
    accept:
        Whether to accept the model.
    proceed:
        How the method should proceed.
    """

    accept: bool
    # TODO change to bool?
    proceed: MethodSignalProceed


class MethodLogger:
    """Log results from a model selection method.

    Attributes
    ----------
    column_width:
        The width of columns when logging.
    column_sep:
        The substring used to separate column values when logging.
    level:
        The logging level.
    logger:
        A logger from the :mod:`logging` module.
    """

    column_width: int = 12
    column_sep: str = " | "

    def __init__(self, level: str = "info"):
        self.logger = logging.getLogger(__name__)
        self.level = level

    def log(self, message, level: str = None) -> None:
        """Log a message.

        Parameters
        ----------
        message:
            The message.
        level:
            The logging level. Defaults to the value defined in the
            constructor.
        """
        if level is None:
            level = self.level
        getattr(self.logger, level)(message)

    def new_selection(self) -> None:
        """Start logging a new model selection."""
        padding = 20
        self.log("-" * padding + "New Selection" + "-" * padding)
        columns = {
            "Predecessor model subspace:ID": "model0",
            "Model subspace:ID": "model",
            "Criterion ID": "crit",
            "Predecessor model criterion": "model0_crit",
            "Model criterion": "model_crit",
            "Criterion difference": "crit_diff",
            "Accept": "accept",
        }
        columns = {
            k: v.ljust(self.column_width)[: self.column_width]
            for k, v in columns.items()
        }
        self.log(self.column_sep.join(columns.values()))

    def new_result(
        self,
        accept,
        criterion,
        model,
        predecessor_model,
        max_id_length: str = 12,
        precision: int = 3,
    ) -> None:
        """Log a model calibration result.

        Parameters
        ----------
        accept:
            Whether the model is accepted.
        criterion:
            The criterion type.
        max_id_length:
            Model and predecessor model IDs are truncated to this length in the
            logged message.
        model:
            The calibrated model.
        predecessor_model:
            The predecessor model.
        precision:
            The number of decimal places to log.
        """
        model_criterion = model.get_criterion(criterion)

        def get_model_id(model: Model) -> str:
            """Get a model ID for logging.

            Parameters
            ----------
            model:
                The model.

            Returns
            -------
            The ID.
            """
            model_subspace_id = model.model_subspace_id or ""
            original_model_id = model.model_id or model.get_hash()
            model_id = model_subspace_id + ":" + original_model_id
            return model_id

        def float_to_str(value: float, precision: int = 3) -> str:
            return f"{value:.{precision}e}"

        if isinstance(predecessor_model, Model):
            predecessor_model_id = get_model_id(predecessor_model)
            predecessor_model_criterion = predecessor_model.get_criterion(
                criterion
            )
            criterion_difference = float_to_str(
                model_criterion - predecessor_model_criterion
            )
            predecessor_model_criterion = float_to_str(
                predecessor_model_criterion
            )
        else:
            criterion_difference = None
            predecessor_model_criterion = None
            predecessor_model_id = predecessor_model

        model_criterion = float_to_str(model_criterion)

        message_parts = [
            predecessor_model_id,
            get_model_id(model),
            criterion.value,
            predecessor_model_criterion,
            model_criterion,
            criterion_difference,
            accept,
        ]
        message = self.column_sep.join(
            [
                str(v).ljust(self.column_width)[: self.column_width]
                for v in message_parts
            ]
        )
        self.log(message)


class MethodCaller:
    """Handle calls to PEtab Select model selection methods.

    Attributes
    ----------
    petab_select_problem:
        The PEtab Select problem.
    candidate_space:
        A :class:`petab_select.CandidateSpace`, used to generate candidate models.
    criterion:
        The criterion by which models will be compared.
    criterion_threshold:
        The minimum improvement in `criterion` that a test model must have to
        be selected. The comparison is made according to the method. For
        example, in `ForwardSelector`, test models are compared to the
        previously selected model.
    calibrated_models:
        All calibrated models of the model selection.
    limit:
        Limit the number of calibrated models. NB: the number of accepted
        models may (likely) be fewer.
    logger:
        A :class:`MethodLogger`, used to log results.
    predecessor_model:
        Specify the predecessor (initial) model for the model selection
        algorithm. If ``None``, then the algorithm will generate an initial
        predecessor model if required.
    user_calibrated_models:
        Supply calibration results for models yourself, as a list of models.
        If a model with the same hash is encountered in the current model
        selection run, and the user-supplied calibrated model has the
        `criterion` value set, the model will not be calibrated again.
    select_first_improvement:
        If ``True``, model selection will terminate as soon as a better model
        is found. If `False`, all candidate models will be tested.
    startpoint_latest_mle:
        If ``True``, one of the startpoints in the multistart optimization
        will be the MLE of the latest model.
    """

    def __init__(
        self,
        petab_select_problem: petab_select.Problem,
        calibrated_models: Models,
        # Arguments/attributes that can simply take the default value here.
        criterion_threshold: float = 0.0,
        limit: int = np.inf,
        # TODO deprecated
        minimize_options: dict = None,
        # TODO deprecated
        model_postprocessor: TYPE_POSTPROCESSOR = None,
        # TODO deprecated
        objective_customizer: Callable = None,
        select_first_improvement: bool = False,
        startpoint_latest_mle: bool = True,
        # Arguments/attributes that should be handled more carefully.
        candidate_space: CandidateSpace = None,
        criterion: Criterion = None,
        # TODO misleading, `Method` here is simply an Enum, not a callable...
        method: Method = None,
        predecessor_model: Model = None,
        # TODO deprecated
        model_to_pypesto_problem_method: Callable[[Any], Problem] = None,
        model_problem_options: dict = None,
        user_calibrated_models: list[Model] = None,
    ):
        """Arguments are used in every `__call__`, unless overridden."""
        self.petab_select_problem = petab_select_problem
        self.calibrated_models = calibrated_models

        self.criterion_threshold = criterion_threshold
        self.limit = limit
        self.predecessor_model = predecessor_model
        self.select_first_improvement = select_first_improvement
        self.startpoint_latest_mle = startpoint_latest_mle

        self.user_calibrated_models = Models()
        if user_calibrated_models is not None:
            self.user_calibrated_models = user_calibrated_models

        self.logger = MethodLogger()

        # TODO deprecated
        old_model_problem_options = {}
        for key, value in [
            ("postprocessor", model_postprocessor),
            (
                "model_to_pypesto_problem_method",
                model_to_pypesto_problem_method,
            ),
            ("minimize_options", minimize_options),
            ("objective_customizer", objective_customizer),
        ]:
            if value is not None:
                old_model_problem_options[key] = value
                self.logger.log(
                    f"Specifying `{key}` as an individual argument is "
                    "deprecated. Please instead specify it within some "
                    "`model_problem_options` dictionary, e.g. "
                    f'`model_problem_options={{"{key}": ...}}`.',
                    level="warning",
                )
        self.model_problem_options = {}
        self.model_problem_options |= old_model_problem_options
        if model_problem_options is not None:
            self.model_problem_options |= model_problem_options

        self.criterion = criterion
        if self.criterion is None:
            self.criterion = self.petab_select_problem.criterion

        # Forbid specification of both a candidate space and a method.
        if candidate_space is not None and method is not None:
            self.logger.log(
                (
                    "Both `candidate_space` and `method` were provided. "
                    "Please only provide one. The method will be ignored here."
                ),
                level="warning",
            )
        # Get method.
        self.method = (
            method
            if method is not None
            else (
                candidate_space.method
                if candidate_space is not None
                else self.petab_select_problem.method
            )
        )
        # Require either a candidate space or a method.
        if candidate_space is None and self.method is None:
            raise ValueError(
                "Please provide one of either `candidate_space` or `method`, "
                "or specify the `method` in the PEtab Select problem."
            )
        # Use candidate space if provided.
        if candidate_space is not None:
            self.candidate_space = candidate_space
            if predecessor_model is not None:
                candidate_space.set_predecessor_model(predecessor_model)
        # Else generate one based on the PEtab Select problem.
        else:
            if predecessor_model is not None:
                self.candidate_space = (
                    self.petab_select_problem.new_candidate_space(
                        method=self.method,
                        predecessor_model=self.predecessor_model,
                    )
                )
            else:
                self.candidate_space = (
                    self.petab_select_problem.new_candidate_space(
                        method=self.method,
                    )
                )
        # May have changed from `None` to `petab_select.VIRTUAL_INITIAL_MODEL`
        self.predecessor_model = self.candidate_space.get_predecessor_model()

    def __call__(self) -> tuple[Model, Models]:
        """Run a single iteration of the model selection method.

        A single iteration here refers to calibration of all candidate models.
        For example, given a predecessor model with 3 estimated parameters,
        with the forward method, a single iteration would involve calibration
        of all models that have both: the same 3 estimated parameters; and 1
        additional estimated parameter.

        Returns
        -------
        A 2-tuple, with the following values:

           1. the predecessor model for the newly calibrated models; and
           2. the newly calibrated models.
        """
        # All calibrated models in this iteration (see second return value).
        self.logger.new_selection()

        iteration = petab_select.ui.start_iteration(
            problem=self.petab_select_problem,
            candidate_space=self.candidate_space,
            limit=self.limit,
            criterion=self.criterion,
            user_calibrated_models=self.user_calibrated_models,
        )

        if not iteration[UNCALIBRATED_MODELS]:
            raise StopIteration("No valid models found.")

        # TODO parallelize calibration (maybe not sensible if
        #      `self.select_first_improvement`)
        calibrated_models = Models()
        for model in iteration[UNCALIBRATED_MODELS]:
            if (
                model.get_criterion(
                    criterion=self.criterion,
                    compute=True,
                    raise_on_failure=False,
                )
                is not None
            ):
                self.logger.log(
                    message=(
                        "Unexpected calibration result already available for "
                        f"model: `{model.get_hash()}`. Skipping "
                        "calibration."
                    ),
                    level="warning",
                )
            else:
                self.new_model_problem(model=model)

            calibrated_models.append(model)
            method_signal = self.handle_calibrated_model(
                model=model,
                predecessor_model=iteration[PREDECESSOR_MODEL],
            )
            if method_signal.proceed == MethodSignalProceed.STOP:
                break

        iteration_results = petab_select.ui.end_iteration(
            problem=self.petab_select_problem,
            candidate_space=iteration[CANDIDATE_SPACE],
            calibrated_models=calibrated_models,
        )

        self.calibrated_models += iteration_results[MODELS]

        return iteration_results[MODELS]

    def handle_calibrated_model(
        self,
        model: Model,
        predecessor_model: Model,
    ) -> MethodSignal:
        """Handle the model selection method, given a new calibrated model.

        Parameters
        ----------
        model:
            The calibrated model.
        predecessor_model:
            The predecessor model.

        Returns
        -------
        A :class:`MethodSignal` that describes the result.
        """
        # Use the predecessor model from `__init__` if an iteration-specific
        # predecessor model was not supplied to `__call__`.
        if predecessor_model is None:
            # May still be `None` after this assignment.
            predecessor_model = self.predecessor_model

        # Default to accepting the model and continuing the method.
        method_signal = MethodSignal(
            accept=True,
            proceed=MethodSignalProceed.CONTINUE,
        )

        # Reject the model if it doesn't improve on the predecessor model.
        if (
            predecessor_model.hash != VIRTUAL_INITIAL_MODEL.hash
            and not self.model1_gt_model0(
                model1=model, model0=predecessor_model
            )
        ):
            method_signal.accept = False

        # Stop the model selection method if it a first improvement is found.
        if self.select_first_improvement and method_signal.accept:
            method_signal.proceed = MethodSignalProceed.STOP

        # TODO allow users to supply an arbitrary constraint function to e.g.:
        #      - quit after 10 accepted models
        #      - reject models that are worse than the current 10 best models

        # Log result
        self.logger.new_result(
            accept=method_signal.accept,
            criterion=self.criterion,
            model=model,
            predecessor_model=predecessor_model,
        )

        return method_signal

    def model1_gt_model0(
        self,
        model1: Model,
        model0: Model,
    ) -> bool:
        """Compare models by criterion.

        Parameters
        ----------
        model1:
            The new model.
        model0:
            The original model.

        Returns
        -------
        ``True``, if `model1` is superior to `model0` by the criterion,
        else ``False``.
        """
        if self.criterion in [
            Criterion.AIC,
            Criterion.AICC,
            Criterion.BIC,
            Criterion.LH,
            Criterion.LLH,
            Criterion.NLLH,
        ]:
            result = petab_select.model.default_compare(
                model0=model0,
                model1=model1,
                criterion=self.criterion,
                criterion_threshold=self.criterion_threshold,
            )
        else:
            raise NotImplementedError(
                f"Model selection criterion: {self.criterion}."
            )
        return result

    def new_model_problem(
        self,
        model: Model,
        valid: bool = True,
        autorun: bool = True,
    ) -> ModelProblem:
        """Create a model problem, usually to calibrate a model.

        Parameters
        ----------
        model:
            The model.
        valid:
            Whether the model should be considered a valid model. If it is
            not valid, it will not be calibrated.
        autorun:
            Whether the model should be calibrated upon creation.

        Returns
        -------
        The model selection problem.
        """
        x_guess = None
        if (
            self.startpoint_latest_mle
            and model.predecessor_model_hash in self.calibrated_models
        ):
            predecessor_model = self.calibrated_models[
                model.predecessor_model_hash
            ]
            if str(model.model_subspace_petab_yaml) != str(
                predecessor_model.model_subspace_petab_yaml
            ):
                raise NotImplementedError(
                    "The PEtab YAML files differ between the model and its "
                    "predecessor model. This may imply different (fixed union "
                    "estimated) parameter sets. Support for this is not yet "
                    "implemented."
                )
            x_guess = {
                **predecessor_model.parameters,
                **predecessor_model.estimated_parameters,
            }

        return ModelProblem(
            model=model,
            criterion=self.criterion,
            valid=valid,
            autorun=autorun,
            x_guess=x_guess,
            **self.model_problem_options,
        )



"""Various methods for estimating the log evidence of a model."""

import logging
from collections.abc import Callable

import numpy as np
from scipy import stats
from scipy.integrate import simpson, trapezoid
from scipy.optimize import minimize_scalar
from scipy.special import logsumexp

from ..C import SIMPSON, STEPPINGSTONE, TRAPEZOID
from ..objective import (
    AggregatedObjective,
    NegLogParameterPriors,
    NegLogPriors,
)
from ..problem import Problem
from ..result import Result
from .diagnostics import geweke_test

logger = logging.getLogger(__name__)


def laplace_approximation_log_evidence(
    problem: Problem, x: np.ndarray
) -> float:
    """
    Compute the log evidence using the Laplace approximation.

    The objective in your `problem` must be a negative log posterior, and support Hessian computation.

    Parameters
    ----------
    problem:
        The problem to compute the log evidence for.
    x:
        The maximum a posteriori estimate at which to compute the log evidence.

    Returns
    -------
    log_evidence: float
    """
    hessian = problem.objective(
        problem.get_reduced_vector(x), sensi_orders=(2,)
    )
    _, log_det = np.linalg.slogdet(hessian)
    log_prop_posterior = -problem.objective(problem.get_reduced_vector(x))
    log_evidence = (
        0.5 * np.log(2 * np.pi) * len(problem.x_free_indices)
        - 0.5 * log_det
        + log_prop_posterior
    )
    return log_evidence


def harmonic_mean_log_evidence(
    result: Result,
    prior_samples: np.ndarray | None = None,
    neg_log_likelihood_fun: Callable | None = None,
) -> float:
    """
    Compute the log evidence using the harmonic mean estimator.

    Stabilized harmonic mean estimator is used if prior samples are provided.
    Newton and Raftery (1994): https://doi.org/10.1111/j.2517-6161.1994.tb01956.x

    Parameters
    ----------
    result:
    prior_samples:
        Samples from the prior distribution. If samples from the prior are provided,
        the stabilized harmonic mean is computed (recommended). Then, the likelihood function must be provided as well.
    neg_log_likelihood_fun: callable
        Function to evaluate the negative log likelihood. Necessary if prior_samples is not `None`.

    Returns
    -------
    log_evidence
    """
    if result.sample_result is None:
        raise ValueError("No samples available. Run sampling first.")

    # compute negative log likelihood from traces
    burn_in = geweke_test(result)
    trace_neglogpost = result.sample_result.trace_neglogpost[0, burn_in:]
    trace_neglogprior = result.sample_result.trace_neglogprior[0, burn_in:]
    neg_log_likelihoods_posterior = trace_neglogpost - trace_neglogprior

    if prior_samples is None:
        # compute harmonic mean from samples
        return -logsumexp(neg_log_likelihoods_posterior) + np.log(
            neg_log_likelihoods_posterior.size
        )

    # compute stabilized harmonic mean
    if prior_samples is not None and neg_log_likelihood_fun is None:
        raise ValueError(
            "you need to provide a likelihood function to evaluate prior samples"
        )

    # compute delta (ratio of prior to posterior samples)
    n_samples_prior = len(prior_samples)
    n_samples_posterior = len(trace_neglogpost)
    delta = n_samples_prior / (n_samples_prior + n_samples_posterior)
    neg_log_likelihoods_prior = np.array(
        [neg_log_likelihood_fun(x) for x in prior_samples]
    )
    log_likelihoods_stack = -np.concatenate(
        [neg_log_likelihoods_prior, neg_log_likelihoods_posterior]
    )

    def _log_evidence_objective(log_p: float):
        # Helper function to compute the log evidence with stabilized harmonic mean
        log_w_i = logsumexp(
            np.stack(
                (
                    log_p * np.ones_like(log_likelihoods_stack),
                    log_likelihoods_stack,
                ),
                axis=1,
            ),
            b=np.array([delta, 1 - delta]),
            axis=1,
        )
        res, sign = logsumexp(
            [
                log_p,
                logsumexp(log_likelihoods_stack - log_w_i)
                - logsumexp(-log_w_i),
            ],
            b=[1, -1],
            return_sign=True,
        )
        return sign * res

    sol = minimize_scalar(_log_evidence_objective)
    return sol.x


def parallel_tempering_log_evidence(
    result: Result,
    method: str = "trapezoid",
    use_all_chains: bool = True,
) -> float | None:
    """Perform thermodynamic integration or steppingstone sampling to estimate the log evidence.

    Thermodynamic integration is performed by integrating the mean log likelihood over the temperatures.
    Errors might come from the samples itself or the numerical integration.
    Steppingstone sampling is a form of importance sampling that uses the maximum likelihood of each temperature.
    It does not require an integration, but can be biased for a small number of temperatures.
    See (Annis et al., 2019), https://doi.org/10.1016/j.jmp.2019.01.005, for more details.

    This should be used with a beta decay temperature schedule and not with the adaptive version of
     parallel tempering sampling as the temperature schedule is not optimal for thermodynamic integration.

    Parameters
    ----------
    result:
        Result object containing the samples.
    method:
        Integration method, either 'trapezoid' or 'simpson' to perform thermodynamic integration
        (uses scipy for integration) or 'steppingstone' to perform steppingstone sampling.
    use_all_chains:
        If True, calculate burn-in for each chain and use the maximal burn-in for all chains for the integration.
        This will fail if not all chains have converged yet.
        Otherwise, use only the converged chains for the integration (might increase the integration error).
    """
    # compute burn in for all chains but the last one (prior only)
    burn_ins = np.zeros(len(result.sample_result.betas), dtype=int)
    for i_chain in range(len(result.sample_result.betas)):
        burn_ins[i_chain] = geweke_test(result, chain_number=i_chain)
    max_burn_in = int(np.max(burn_ins))

    if max_burn_in >= result.sample_result.trace_x.shape[1]:
        logger.warning(
            f"At least {np.sum(burn_ins >= result.sample_result.trace_x.shape[1])} chains seem to not have "
            f"converged yet. You may want to use a larger number of samples."
        )
        if use_all_chains:
            raise ValueError(
                "Not all chains have converged yet. You may want to use a larger number of samples, "
                "or try ´use_all_chains=False´, which might increase the integration error."
            )

    if use_all_chains:
        # estimate mean of log likelihood for each beta
        trace_loglike = (
            result.sample_result.trace_neglogprior[::-1, max_burn_in:]
            - result.sample_result.trace_neglogpost[::-1, max_burn_in:]
        )
        mean_loglike_per_beta = np.mean(trace_loglike, axis=1)
        temps = result.sample_result.betas[::-1]
    else:
        # estimate mean of log likelihood for each beta if chain has converged
        mean_loglike_per_beta = []
        trace_loglike = []
        temps = []
        for i_chain in reversed(range(len(result.sample_result.betas))):
            if burn_ins[i_chain] < result.sample_result.trace_x.shape[1]:
                # save temperature-chain as it is converged
                temps.append(result.sample_result.betas[i_chain])
                # calculate mean log likelihood for each beta
                trace_loglike_i = (
                    result.sample_result.trace_neglogprior[
                        i_chain, burn_ins[i_chain] :
                    ]
                    - result.sample_result.trace_neglogpost[
                        i_chain, burn_ins[i_chain] :
                    ]
                )
                trace_loglike.append(trace_loglike_i)
                mean_loglike_per_beta.append(np.mean(trace_loglike_i))

    if method == TRAPEZOID:
        log_evidence = trapezoid(
            # integrate from low to high temperature
            y=mean_loglike_per_beta,
            x=temps,
        )
    elif method == SIMPSON:
        log_evidence = simpson(
            # integrate from low to high temperature
            y=mean_loglike_per_beta,
            x=temps,
        )
    elif method == STEPPINGSTONE:
        log_evidence = steppingstone(temps=temps, trace_loglike=trace_loglike)
    else:
        raise ValueError(
            f"Unknown method {method}. Choose 'trapezoid', 'simpson' for thermodynamic integration or ",
            "'steppingstone' for steppingstone sampling.",
        )

    return log_evidence


def steppingstone(temps: np.ndarray, trace_loglike: np.ndarray) -> float:
    """Perform steppingstone sampling to estimate the log evidence.

    Implementation based on  Annis et al. (2019): https://doi.org/10.1016/j.jmp.2019.01.005.

    Parameters
    ----------
    temps:
        Temperature values.
    trace_loglike:
        Log likelihood values for each temperature.
    """
    from scipy.special import logsumexp

    ss_log_evidences = np.zeros(len(temps) - 1)
    for t_i in range(1, len(temps)):
        # we use the maximum likelihood times the temperature difference to stabilize the logsumexp
        # original formulation uses only the maximum likelihood, this is equivalent
        ss_log_evidences[t_i - 1] = logsumexp(
            trace_loglike[t_i - 1] * (temps[t_i] - temps[t_i - 1])
        ) - np.log(trace_loglike[t_i - 1].size)
    log_evidence = np.sum(ss_log_evidences)
    return log_evidence


def bridge_sampling_log_evidence(
    result: Result,
    n_posterior_samples_init: int | None = None,
    initial_guess_log_evidence: float | None = None,
    max_iter: int = 1000,
    tol: float = 1e-6,
) -> float:
    """
    Compute the log evidence using bridge sampling.

    Based on "A Tutorial on Bridge Sampling" by Gronau et al. (2017): https://doi.org/10.1016/j.jmp.2017.09.005.
    Using the optimal bridge function by Meng and Wong (1996) which minimises the relative mean-squared error.
    Proposal function is calibrated using posterior samples, which are not used for the final bridge estimate
    (as this may result in an underestimation of the marginal likelihood, see Overstall and Forster (2010)).

    Parameters
    ----------
    result:
        The pyPESTO result object with filled sample result.
    n_posterior_samples_init:
        Number of samples used to calibrate the proposal function. By default, half of the posterior samples are used.
    initial_guess_log_evidence:
        Initial guess for the log evidence. By default, the Laplace approximation is used to compute the initial guess.
    max_iter:
        Maximum number of iterations. Default is 1000.
    tol:
        Tolerance for convergence. Default is 1e-6.


    Returns
    -------
    log_evidence
    """
    if result.sample_result is None:
        raise ValueError("No samples available. Run sampling first.")
    if not isinstance(result.problem.objective, AggregatedObjective):
        raise ValueError("Objective must be an AggregatedObjective.")

    # use Laplace approximation to get initial guess for p(y)
    if initial_guess_log_evidence is None:
        initial_guess_log_evidence = laplace_approximation_log_evidence(
            problem=result.problem, x=result.optimize_result.x[0]
        )
    # extract posterior samples
    burn_in = geweke_test(result)
    posterior_samples = result.sample_result.trace_x[0, burn_in:]

    # build proposal function from posterior samples
    if n_posterior_samples_init is None:
        n_posterior_samples_init = int(posterior_samples.shape[0] * 0.5)
    # randomly select samples for calibration
    calibration_index = np.random.choice(
        np.arange(posterior_samples.shape[0]),
        n_posterior_samples_init,
        replace=False,
    )
    samples_calibration = posterior_samples[calibration_index]
    # remove calibration samples from posterior samples
    posterior_samples = posterior_samples[
        [
            j
            for j in range(posterior_samples.shape[0])
            if j not in calibration_index
        ]
    ]
    # generate proposal samples and define proposal function
    n_proposal_samples = posterior_samples.shape[0]
    posterior_mean = np.mean(samples_calibration, axis=0)
    posterior_cov = np.cov(samples_calibration.T)
    # if covariance matrix is not positive definite (numerically), use diagonal covariance matrix only
    try:
        # proposal density function
        log_proposal_fun = stats.multivariate_normal(
            mean=posterior_mean, cov=posterior_cov
        ).logpdf
    except np.linalg.LinAlgError:
        posterior_cov = np.diag(np.diag(posterior_cov))
        log_proposal_fun = stats.multivariate_normal(
            mean=posterior_mean, cov=posterior_cov
        ).logpdf

    # generate proposal samples
    if posterior_cov.size == 1:
        # univariate case
        proposal_samples = np.random.normal(
            loc=posterior_mean,
            scale=np.sqrt(posterior_cov),
            size=n_proposal_samples,
        )
        proposal_samples = proposal_samples.reshape(-1, 1)
    else:
        # multivariate case
        proposal_samples = np.random.multivariate_normal(
            mean=posterior_mean, cov=posterior_cov, size=n_proposal_samples
        )

    # Compute the weights for the bridge sampling estimate
    log_s1 = np.log(
        posterior_samples.shape[0]
        / (posterior_samples.shape[0] + n_proposal_samples)
    )
    log_s2 = np.log(
        n_proposal_samples / (posterior_samples.shape[0] + n_proposal_samples)
    )

    # Start with the initial guess for p(y)
    log_p_y = initial_guess_log_evidence

    # Compute the log-likelihood, log-prior, and log-proposal for the posterior and proposal samples
    # assumes that the objective function is the negative log-likelihood + negative log-prior

    # get index of prior in the objective function
    likelihood_fun_indices = []
    for i, obj in enumerate(result.problem.objective._objectives):
        if not isinstance(obj, NegLogParameterPriors) and not isinstance(
            obj, NegLogPriors
        ):
            likelihood_fun_indices.append(i)

    def log_likelihood_fun(x_array):
        return np.array(
            [
                np.sum(
                    [
                        -obj(
                            result.problem.get_full_vector(
                                x=x, x_fixed_vals=result.problem.x_fixed_vals
                            )
                        )
                        for obj_i, obj in enumerate(
                            result.problem.objective._objectives
                        )
                        if obj_i in likelihood_fun_indices
                    ]
                )
                for x in x_array
            ]
        )

    def log_prior_fun(x_array):
        return np.array(
            [
                np.sum(
                    [
                        -obj(
                            result.problem.get_full_vector(
                                x=x, x_fixed_vals=result.problem.x_fixed_vals
                            )
                        )
                        for obj_i, obj in enumerate(
                            result.problem.objective._objectives
                        )
                        if obj_i not in likelihood_fun_indices
                    ]
                )
                for x in x_array
            ]
        )

    log_likelihood_posterior = log_likelihood_fun(posterior_samples)
    log_prior_posterior = log_prior_fun(posterior_samples)
    log_proposal_posterior = log_proposal_fun(posterior_samples)

    log_likelihood_proposal = log_likelihood_fun(proposal_samples)
    log_prior_proposal = log_prior_fun(proposal_samples)
    log_proposal_proposal = log_proposal_fun(proposal_samples)

    log_h_posterior_1 = log_s1 + log_likelihood_posterior + log_prior_posterior
    log_h_proposal_1 = log_s1 + log_likelihood_proposal + log_prior_proposal
    for i in range(max_iter):
        # Compute h(θ) for posterior samples
        log_h_posterior_2 = log_s2 + log_p_y + log_proposal_posterior
        log_h_posterior = logsumexp([log_h_posterior_1, log_h_posterior_2])

        # Compute h(θ) for proposal samples
        log_h_proposal_2 = log_s2 + log_p_y + log_proposal_proposal
        log_h_proposal = logsumexp([log_h_proposal_1, log_h_proposal_2])

        # Calculate the numerator and denominator for the bridge sampling estimate
        temp = log_likelihood_proposal + log_prior_proposal + log_h_proposal
        log_numerator = logsumexp(temp) - np.log(
            temp.size
        )  # compute mean in log space
        temp = log_proposal_posterior + log_h_posterior
        log_denominator = logsumexp(temp) - np.log(
            temp.size
        )  # compute mean in log space

        # Update p(y)
        log_p_y_new = log_numerator - log_denominator

        # Check for convergence
        if abs(log_p_y_new - log_p_y) < tol:
            break

        log_p_y = log_p_y_new

        if i == max_iter - 1:
            logger.warning(
                "Bridge sampling did not converge in the given number of iterations."
            )

    return log_p_y



import logging

from ...C import InnerParameterType
from ..base_parameter import InnerParameter

logger = logging.getLogger(__name__)


class RelativeInnerParameter(InnerParameter):
    """A relative (inner) parameter of the relative hierarchical optimization problem.

    Attributes
    ----------
    observable_ids:
        A list of IDs of the observables that the relative inner parameter is connected to.
        Can be a single or multiple observables.
    observable_indices:
        A list of indices of the observables that the relative inner parameter is connected to.
    """

    def __init__(
        self,
        *args,
        observable_ids: list[str] = None,
        observable_indices: list[int] = None,
        **kwargs,
    ):
        """Construct.

        Parameters
        ----------
        See class attributes.
        """
        super().__init__(*args, **kwargs)
        if self.inner_parameter_type not in [
            InnerParameterType.SCALING,
            InnerParameterType.OFFSET,
            InnerParameterType.SIGMA,
        ]:
            raise ValueError(
                "For the RelativeParameter class, the parameter type has to be scaling, offset or sigma."
            )

        self.observable_ids = observable_ids
        self.observable_indices = observable_indices
